import torch as _torch
import torch.nn as _nn

from tamm.ao import layers as _ao_layers
from tamm.ao.arch_optimizers import common as _common
from tamm.utils import OptionalBool as _OptionalBool


class KVQuantArchOptimizer(_common.ArchOptimizer):
    """
    An :class:`.ArchOptimizer` that applies "fake" activation quantization to the keys
    and values of attention layers.  This helps simulate the impact of KV-cache
    quantization.

    Specifically, the :meth:`.optimize` method of this class inserts :obj:`.KVQuantizer`
    layers immediately prior to the ``kv_cacher`` in :obj:`.TransformerAttention`
    layers.  The quantization uses :class:`.FakeQuantize` layers and
    :obj:`.SimpleEMAMinMaxObserver` observers with default arguments.

    Args:
        freeze_qparams (:obj:`bool`, optional): A flag for freezing the quantization
            parameters (such as scales and zero points).  If ``False`` (the default),
            then observers from the key and value quantizers are enabled, meaning that
            they update the qparams during each forward pass in training mode (the
            params remain frozen when in eval mode).
        cast_dtype (:obj:`torch.dtype`, optional): A ``cast_dtype`` argument for
            the key and value :class:`.FakeQuantize` layers.  Defaults to ``float32``.
        skip_quantize_for_first_token(:obj:`bool`, optional): Set to ``True`` to quantize
            kv cacher while skipping the quantization for first token logit calculation
            but still writing quantized values to the cache.
            Defaults to ``False``.
    """

    freeze_qparams: bool = False
    cast_dtype: _torch.dtype = _torch.float32
    skip_quantize_for_first_token: bool = False

    def _optimize_impl(
        self,
        model: _nn.Module,
        *,
        pretrained: "_OptionalBool" = _OptionalBool.NOTSET,
    ) -> None:
        """
        Modifies ``TransformerAttention`` module based on if ``skip_quantize_for_first_token`` is:
        - ``True``: Wraps the kv_cacher layer to include special quantization: skipping the
            quantization for first token logit calculation but still writing quantized values to the cache.
        - ``False``: Inserts kv_quantizer layer above kv_cacher layer.
        """
        # pylint: disable-next=import-outside-toplevel
        from tamm import layers

        for layer in model.modules():
            if not isinstance(layer, layers.TransformerAttention):
                continue

            if self.skip_quantize_for_first_token:
                layer["kv_cacher"] = self._create_quantizing_kv_cacher_layer(
                    layer["kv_cacher"]
                )
            else:
                kv_cacher_idx = list(layer).index(layer["kv_cacher"])
                quantizer = self._create_quantizer_layer()
                layer.insert(kv_cacher_idx, quantizer, name="kv_quantizer")

    def _create_quantizer_layer(self):
        key_quantizer = self._create_fake_quantizer()
        value_quantizer = self._create_fake_quantizer()
        return _ao_layers.KVQuantizer(
            key_quantizer=key_quantizer, value_quantizer=value_quantizer
        )

    def _create_fake_quantizer(self):
        """Creates a FakeQuantize layer."""
        enable_observer = "disabled" if self.freeze_qparams else "only_training"
        return _ao_layers.FakeQuantize(
            _ao_layers.SimpleEMAMinMaxObserver(),
            enable_observer=enable_observer,
            cast_dtype=self.cast_dtype,
        )

    def _create_quantizing_kv_cacher_layer(self, kv_cacher):
        """Wraps the kv_cacher layer to include quantization, skipping the quantization for first
        token logit calculation but still writes quantized values to the cache."""
        key_quantizer = self._create_fake_quantizer()
        value_quantizer = self._create_fake_quantizer()
        return _ao_layers.QuantizingKVCacher(key_quantizer, value_quantizer, kv_cacher)



import base64 as _base64
import enum as _enum
import io as _io
import logging as _logging
import re as _re
from typing import Optional as _Optional
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.typing import PathLike as _PathLike
from tamm.utils.uri import _is_uri, _URIHandler

# pylint: disable=import-outside-toplevel


_logger = _logging.getLogger(__name__)


class LoadingModule(_enum.Enum):
    """
    An enum for specifying the Python module that loads images in
    :func:`load_image`.
    """

    #: Use ``torchvision`` but fall back to ``PIL`` if ``PIL`` is
    #: available and ``torchvision`` is not.
    AUTO = 1

    #: Use ``torchvision``.
    TORCHVISION = 2

    #: Use ``PIL``.
    PIL = 3


def load_image(
    uri: _Union[str, _PathLike],
    mode: _Optional[str] = "RGB",
    loading_module: _Union[LoadingModule, str] = LoadingModule.AUTO,
) -> _torch.Tensor:
    """
    Loads and returns an image as a :obj:`torch.Tensor` in channels-first format.

    Args:
        uri (:obj:`str`): A string representation of the image.  This can
            be a local filepath, a URI, or a base64-encoded image string.
        mode (:obj:`str`): The image mode for the result.  Defaults to RGB.
        loading_module (:obj:`LoadingModule` or :obj:`str`): Specification of
            the Python module for loading the image.  Defaults to ``AUTO``.

    Returns:
        A :obj:`torch.Tensor` of pixel values in channels-first format.
    """

    if isinstance(uri, str):
        is_base64_encoded = _re.match("data:image/.*;base64,", uri)
        if is_base64_encoded is not None:
            data = uri[is_base64_encoded.end() :]
            return _load_base64_encoded_image(
                data, mode=mode, loading_module=loading_module
            )

    if _is_uri(uri):
        local_path = _URIHandler().map_to_local(uri)
    else:
        local_path = uri

    return _load_image_from_local_file(
        local_path, mode=mode, loading_module=loading_module
    )


def _load_image_from_local_file(
    filepath: _Union[str, _PathLike],
    *,
    mode: _Optional[str],
    loading_module: _Union[str, LoadingModule],
) -> _torch.Tensor:
    loading_module = _helpers.get_enum_member_from_name(LoadingModule, loading_module)

    if loading_module is LoadingModule.TORCHVISION:
        return _decode_image_torchvision(filepath, mode=mode)

    if loading_module is LoadingModule.PIL:
        return _decode_image_pil(filepath, mode=mode)

    return _decode_image_auto(filepath, mode=mode)


def _decode_image_auto(filepath, *, mode):
    try:
        return _decode_image_torchvision(filepath, mode=mode)
    except Exception as e:
        _logger.debug(f"Caught exception when loading image with torchvision: {e}")
        try:
            return _decode_image_pil(filepath, mode=mode)
        except Exception as e2:
            raise e2 from e


def _decode_image_torchvision(filepath, *, mode):
    import torchvision.io

    filepath = str(filepath)  # important older torch versions

    if mode is None:
        mode = "UNCHANGED"
    mode = _helpers.get_enum_member_from_name(torchvision.io.ImageReadMode, mode)

    return torchvision.io.read_image(filepath, mode=mode)


def _decode_image_pil(filepath, *, mode):
    import PIL.Image

    image = PIL.Image.open(filepath)
    if mode is not None:
        image = image.convert(mode)
    return _convert_pil_to_torch(image)


def _load_base64_encoded_image(
    data: str,
    mode: _Optional[str],
    loading_module: _Union[LoadingModule, str] = LoadingModule.AUTO,
) -> _torch.Tensor:
    """Loads an image from a base64-encoded string."""

    loading_module = _helpers.get_enum_member_from_name(LoadingModule, loading_module)

    decoded_bytes = _base64.b64decode(data)

    if loading_module is LoadingModule.TORCHVISION:
        return _load_base64_encoded_image_torchvision(decoded_bytes, mode=mode)

    if loading_module is LoadingModule.PIL:
        return _load_base64_encoded_image_pil(decoded_bytes, mode=mode)

    return _load_base64_encoded_image_auto(decoded_bytes, mode=mode)


def _load_base64_encoded_image_auto(data: bytes, *, mode) -> _torch.Tensor:
    try:
        return _load_base64_encoded_image_torchvision(data, mode=mode)
    except Exception as e:
        _logger.debug(f"Caught exception when loading image with torchvision: {e}")

        try:
            return _load_base64_encoded_image_pil(data, mode=mode)
        except Exception as e2:
            raise e2 from e


def _load_base64_encoded_image_torchvision(data: bytes, *, mode) -> _torch.Tensor:
    import torchvision.io

    if mode is None:
        mode = "UNCHANGED"
    mode = _helpers.get_enum_member_from_name(torchvision.io.ImageReadMode, mode)

    data = _torch.frombuffer(data, dtype=_torch.uint8)
    return torchvision.io.decode_image(data, mode=mode)


def _load_base64_encoded_image_pil(data: bytes, *, mode) -> _torch.Tensor:
    import PIL.Image

    bytes_io = _io.BytesIO(data)
    image = PIL.Image.open(bytes_io)
    if mode is not None:
        image = image.convert(mode)
    return _convert_pil_to_torch(image)


def _convert_pil_to_torch(image) -> _torch.Tensor:
    """
    Converts a PIL Image to a torch Tensor (without assuming a torchvision dependency)
    """
    import numpy as np

    is_binary = image.mode == "1"

    image = np.array(image)
    if image.ndim == 2:
        image = image[..., None]
    image = np.moveaxis(image, -1, 0)  # make channels-first

    if is_binary or image.dtype == np.uint8:
        dtype = _torch.uint8
    else:
        dtype = None
    image = _torch.tensor(image, dtype=dtype)

    if is_binary:
        image *= 255

    return image



"""
Command for cleaning cache or temporary directories
"""

import shutil as _shutil

from tamm._cli.common import argument, command
from tamm.utils import user_dir_utils


@command("clean", help_text="Clean cache or temporary directories")
@argument(
    "entity",
    nargs="?",
    choices=["cache", "tempdir", "all"],
    default="all",
    help="The location(s) to clean",
)
def clean(entity):
    paths = {}
    if entity == "cache":
        paths["cache"] = user_dir_utils.get_tamm_cache_dir()
    elif entity == "tempdir":
        paths["tempdir"] = user_dir_utils.get_tamm_tmp_dir()
    elif entity == "all":
        paths["tempdir"] = user_dir_utils.get_tamm_tmp_dir()
        paths["cache"] = user_dir_utils.get_tamm_cache_dir()
    else:
        raise ValueError(f"Unknown entity type '{entity}'. Cannot clean this path.")

    for name, path in paths.items():
        print(f"Cleaning '{name}': {path}")
        _shutil.rmtree(path, ignore_errors=False)
        path.mkdir(parents=True, exist_ok=True)



from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

from tamm._adapters_v1.layer_adapters.common import AdapterMode as _AdapterMode
from tamm._adapters_v1.layer_adapters.common import (
    CompositeInputTransform as _CompositeInputTransform,
)
from tamm._adapters_v1.layer_adapters.common import (
    CompositeOutputTransform as _CompositeOutputTransform,
)
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm.utils._torch_compatibility import _pytree_average


class AveragedInputTransforms(_CompositeInputTransform):
    """
    A composite :class:`LayerAdapter` that averages together input transformations from
    multiple child adapters.  The behavior is equivalent to the following:

    .. code-block:: python

        transformed_inputs = [
            adapter(TRANSFORM_INPUTS, args=args, kwargs=kwargs)
            for adapter in adapters.values()
        ]
        args, kwargs = average(transformed_inputs, weights=weights)
        outputs = wrapped(args, kwargs)

    The transformed inputs returned by each adapter may be a container object as long
    as the structure of the object is identical across adapters.  In this case, the
    average applies to all floating point tensors in the container and the result's
    other values come from the transformed inputs of the first adapter.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.
        weights (:obj:`dict` that maps :obj:`str` to :obj:`float`, optional): A
            dictionary that maps child adapter IDs to corresponding weights for
            averaging transformed inputs.  Defaults to ``None``, which results in no
            weighting.
    """

    def __init__(
        self,
        adapters: _Dict[str, _LayerAdapter],
        weights: _Optional[_Dict[str, float]] = None,
    ):
        super().__init__(adapters=adapters)
        self.weights = weights

    def _transform_inputs(self, *, args: _Any, kwargs: _Any) -> _Any:
        transformed_inputs = [
            adapter(mode=_AdapterMode.TRANSFORM_INPUTS, args=args, kwargs=kwargs)
            for adapter in self._child_adapters.values()
        ]
        if self.weights is not None:
            weights = [self.weights[key] for key in self._child_adapters]
        else:
            weights = None
        return _pytree_average(*transformed_inputs, weights=weights)


class AveragedOutputTransforms(_CompositeOutputTransform):
    """
    A composite :class:`LayerAdapter` that averages together output transformations from
    multiple child adapters.  The behavior is equivalent to the following:

    .. code-block:: python

        outputs = wrapped(args, kwargs)
        transformed_outputs = [
            adapter(
                TRANSFORM_OUTPUTS,
                args=args,
                kwargs=kwargs,
                transformed_args=args,
                transformed_kwargs=kwargs,
                outputs=outputs,
            )
            for adapter in adapters.values()
        ]
        outputs = average(transformed_outputs, weights=weights)

    The transformed outputs returned by each adapter may be a container object as long
    as the structure of the object is identical across adapters.  In this case, the
    average applies to all floating point tensors in the container and the result's
    other values come from the transformed outputs of the first adapter.

    Args:
        adapters (:obj:`dict` that maps :obj:`str` to :obj:`LayerAdapter`): A dictionary
            that maps child adapter IDs to corresponding adapters.
        weights (:obj:`dict` that maps :obj:`str` to :obj:`float`, optional): A
            dictionary that maps child adapter IDs to corresponding weights for
            averaging transformed outputs.  Defaults to ``None``, which results in no
            weighting.
    """

    def __init__(
        self,
        adapters: _Dict[str, _LayerAdapter],
        weights: _Optional[_Dict[str, float]] = None,
    ):
        super().__init__(adapters=adapters)
        self.weights = weights

    def _transform_outputs(  # pylint: disable=unused-argument
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _Any,
    ) -> _Any:
        outputs = [
            adapter(
                mode=_AdapterMode.TRANSFORM_OUTPUTS,
                outputs=outputs,
                args=args,
                kwargs=kwargs,
                transformed_args=transformed_args,
                transformed_kwargs=transformed_kwargs,
            )
            for adapter in self._child_adapters.values()
        ]
        if self.weights is not None:
            weights = [self.weights[key] for key in self._child_adapters]
        else:
            weights = None
        return _pytree_average(*outputs, weights=weights)



import contextlib as _contextlib
from contextvars import ContextVar
from typing import Union

_INITIALIZING_PLUGIN_MODULE_NAME = ContextVar[Union[str, None]](
    "initializing_plugin_module_name", default=None
)


def get_initializing_plugin_module_name() -> Union[str, None]:
    """
    Returns the module name of the plugin that is currently initializing.
    Returns ``None`` if no plugin is currently initializing.
    """
    return _INITIALIZING_PLUGIN_MODULE_NAME.get()


@_contextlib.contextmanager
def plugin_initialization_context(module_name: str):
    """
    A context manager for initializing |tamm| plugins, which records the
    module name of the initializing plugin.

    Args:
        module_name (:obj:`str`): The module name of the initializing plugin.
    """
    token = _INITIALIZING_PLUGIN_MODULE_NAME.set(module_name)
    try:
        yield
    finally:
        _INITIALIZING_PLUGIN_MODULE_NAME.reset(token)



from tamm._plugin.cli import register_cli_commands
from tamm._plugin.core_import_callback import execute_core_import_callbacks
from tamm._plugin.core_transformers_import_callback import (
    execute_core_transformers_import_callbacks,
)
from tamm._plugin.utils import (
    discover_and_retrieve_plugins,
    import_discovered_plugins,
    import_named_plugin,
)



import dataclasses as _dataclasses
from typing import TYPE_CHECKING, Optional

from tamm import _helpers
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin

if TYPE_CHECKING:
    from tamm.models.common import ModelConfig


@_dataclasses.dataclass
class PublishedModelConfig(  # type: ignore[call-arg]
    _JSONSerializableMixin, json_namespace="model_repo"
):
    """
    Use the following :class:`tamm.models.AFMText`
    example to create a :class:`.PublishedModelConfig` (same process applies to
    any model supported by |tamm| and |tamm|-plugins):

    .. code-block:: python

        import tamm.models
        from tamm import adapters as _adapters
        import tamm.utils.json as tamm_json
        from tamm.model_repo import PublishedModelConfig

        config = tamm.models.AFMText.Config()
        # fill in fields by modifying the example below
        config.num_layers = 8
        config.adapters = {"my_lora_adapter":
            _adapters.LoRAModelAdapter(
            rank=rank,
            alpha=alpha,
            adapt_attention_queries=adapt_q,
            adapt_attention_keys=adapt_k,
            adapt_attention_values=adapt_v,
            adapt_attention_outputs=True,
            adapt_feed_forward_hidden_states=True,
            adapt_feed_forward_outputs=True,
            dropout_p=lora_dropout_p,
            pretrained_path=lora_pretrained_path,
        )}

        # PublishedModelConfig which includes model ID,
        # description and deprecation info are required for publishing.
        published_config = PublishedModelConfig(
                model_config=config,
                model_id="model_xyz",
                description="model_xyz has 1T parameters, 8 layers, and LoRA adapter",
                # the following 2 attributes are optional
                is_deprecated=True,
                replacement_model_id="model_superseding_xyz",
            )

        with open("config.json", "w") as f:
            tamm_json.dump(published_config, f)
        # share config.json with people who want to use this model
    """

    model_config: "ModelConfig"
    model_id: str
    is_deprecated: bool = False
    replacement_model_id: Optional[str] = None
    description: str = ""

    def __lt__(self, other):
        return self.model_id < other.model_id

    def _to_json_dict_impl(self):
        return _helpers.dataclass_to_dict(self, omit_defaults=True)

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict: dict):
        return _helpers.dataclass_init_drop_missing_keys(raw_dict, dataclass_type=cls)



import abc as _abc
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.typing import OptionalTensor as _OptionalTensor
from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class TokenMetadata:
    """
    A dataclass that holds segment ids and positions for transformer tokens.  The
    tensors for keys and values may have longer sequence lengths than for
    those for queries (if using a KV cache).
    """

    q_segment_ids: _torch.Tensor = None
    """
    Segment IDs for self-attention query tokens.  Tokens only attend to each other when
    the q segment ID matches the kv segment ID.
    """

    kv_segment_ids: _torch.Tensor = None
    """Segment IDs for self-attention key and value tokens."""

    q_positions: _torch.Tensor = None
    """Positions for self-attention query tokens."""

    kv_positions: _torch.Tensor = None
    """Positions for self-attention key and value tokens."""

    q_token_types: _torch.Tensor = None
    """Types for self-attention query tokens (this can be arbitrary metadata)."""

    kv_token_types: _torch.Tensor = None
    """Types for self-attention key and value tokens."""

    aux_metadata: _Dict[str, _Any] = None
    """Auxiliary metadata for extending model behavior."""

    def __post_init__(self):
        if self.q_segment_ids is not None:
            if not _torch.is_tensor(self.q_segment_ids) or self.q_segment_ids.ndim != 2:
                raise ValueError("q_segment_ids must be a 2d tensor")
        if self.kv_segment_ids is not None:
            if (
                not _torch.is_tensor(self.kv_segment_ids)
                or self.kv_segment_ids.ndim != 2
            ):
                raise ValueError("kv_segment_ids must be a 2d tensor")
        if self.q_positions is not None:
            if not _torch.is_tensor(self.q_positions) or self.q_positions.ndim != 2:
                raise ValueError("q_positions must be a 2d tensor")
        if self.kv_positions is not None:
            if not _torch.is_tensor(self.kv_positions) or self.kv_positions.ndim != 2:
                raise ValueError("kv_positions must be a 2d tensor")

        if self.aux_metadata is None:
            self.aux_metadata = {}
        if not isinstance(self.aux_metadata, dict):
            raise ValueError("aux_metadata must be a dictionary")


class TokenMetadataLogic(_torch.nn.Module, _abc.ABC):
    """
    A base class for layers that determine the token metadata (segment ids, positions,
    token types) in a :obj:`TransformerStack`.
    """

    def __init__(self):
        super().__init__()

    def forward(
        self,
        *,
        embeddings: _torch.Tensor,
        segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView] = None,
        aux_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> TokenMetadata:
        """
        Populates a :obj:`TokenMetadata` from :class:`TransformerStack` inputs.
        The layer computes default segment IDs, positions, and token types as needed.
        It also reads cached segment ids and positions from the cache when ``kv_cache``
        is not ``None``.

        Args:
            embeddings (:obj:`torch.Tensor`): The embeddings input to the :class:`TransformerStack`.
            segment_ids (:obj:`torch.Tensor`): An integer tensor with shape
                ``(batch_size, seq_len)`` that controls the attention mask for
                self attention layers.  For sequence ``i``, if
                ``segment_ids[i, j] != segment_ids[i, k]``, then tokens ``j`` and
                ``k`` do not attend to each other.
            positions (:obj:`torch.Tensor`, optional): An optional integer tensor with
                shape ``(batch_size, seq_len)`` that specifies the position of each
                token.  Defaults to ``None``.
            token_types (:obj:`torch.Tensor`, optional): An optional integer tensor with shape
                ``(batch_size, seq_len)`` that specifies the type of each token.  Defaults to
                ``None``.
            kv_cache (:obj:`BaseKVCacheView`, optional): An optional KV cache view
                with cached keys and values, etc.
            aux_metadata (:obj:`dict`): A dictionary of optional auxiliary inputs for
                extending model behavior.

        Returns:
            The newly created :obj:`TokenMetadata`.
        """
        default_segment_ids = None
        if segment_ids is None:
            default_segment_ids = _torch.ones(
                embeddings.shape[:-1], device=embeddings.device, dtype=_torch.int64
            )

        if kv_cache is not None:
            return self._from_kv_cache(
                kv_cache=kv_cache,
                segment_ids=segment_ids,
                default_segment_ids=default_segment_ids,
                positions=positions,
                token_types=token_types,
                aux_metadata=aux_metadata,
            )

        return self._from_q_values(
            segment_ids=segment_ids,
            default_segment_ids=default_segment_ids,
            positions=positions,
            token_types=token_types,
            aux_metadata=aux_metadata,
        )

    @_abc.abstractmethod
    def _from_q_values(
        self,
        segment_ids: _OptionalTensor = None,
        default_segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        aux_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> TokenMetadata:
        """
        Factory method for creating segments and positions without a KV cache.  In this
        case q and kv tensors have the same length.  If positions is None, this defaults
        to 0, 1, 2, ... for each segment of each sequence.
        """

    @_abc.abstractmethod
    def _from_kv_cache(
        self,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView],
        *,
        segment_ids: _OptionalTensor = None,
        default_segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        aux_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> TokenMetadata:
        """
        Factory method for creating segments and positions with a KV cache.  In this
        case kv tensors may have a longer length than k tensors.
        """



from typing import Optional, Union

import torch as _torch

from tamm.context_vars._freeze_params import get_default_freeze_params_flag
from tamm.context_vars._pretrained import get_default_pretrained_flag
from tamm.utils._torch_compatibility import _get_default_device
from tamm.utils.optional_bool import OptionalBool


def resolve_pretrained_flag(pretrained_hint=OptionalBool.NOTSET) -> "OptionalBool":
    """
    Determine ``pretrained`` flag based on the environment.
    """
    pretrained_hint = OptionalBool(pretrained_hint)
    if pretrained_hint != OptionalBool.NOTSET:
        return pretrained_hint
    return get_default_pretrained_flag()


def resolve_freeze_params(hint=OptionalBool.NOTSET) -> "OptionalBool":
    """
    Determine ``freeze_params`` flag based on the environment.
    """
    hint = OptionalBool(hint)
    if hint != OptionalBool.NOTSET:
        return hint
    return get_default_freeze_params_flag()


def resolve_device(
    device_hint: Optional[Union[str, "_torch.device"]] = None
) -> _torch.device:
    """
    Determine an explicit ``torch.device`` based on the environment.

    Args:
        device_hint: an explicit device hint

    Returns: torch.device

    """
    if isinstance(device_hint, str):
        device_hint = _torch.device(device_hint)

    if device_hint is not None:
        return device_hint
    return _get_default_device()



"""
tamm.layers
-----------
"""

from tamm.layers import common, functional, init, mixture_of_experts, torch_nn
from tamm.layers.activation import GEGLU, GELU, GLU, QuickGELU, ReLU, SiLU, SwiGLU, Tanh
from tamm.layers.attention import (
    KVReuseTransformerAttention,
    QKNorm,
    QKVLinear,
    RoPETransform,
    ScaledDotProductAttention,
    TransformerAttention,
)
from tamm.layers.basic import (
    ChannelsFirstToLast,
    ChannelsLastToFirst,
    Index,
    Interpolation,
    Map,
    Mean,
    PadToMultiple,
    SelectByKey,
    SoftCap,
    Sum,
    Union,
)
from tamm.layers.convolution import (
    Conv1d,
    Conv2d,
    ResNetBlock,
    ResNetStage,
    SqueezeExcitation,
)
from tamm.layers.decoding import KVCacher
from tamm.layers.dropout import Dropout
from tamm.layers.embedding import (
    ConcatEmbedding,
    ConstantEmbedding,
    ConvEmbedding,
    Embedding,
    LowRankFactorizedEmbedding,
    UnionEmbedding,
)
from tamm.layers.feed_forward import FeedForward, TransformerFeedForward
from tamm.layers.lambda_layer import Lambda
from tamm.layers.linear import (
    FusedMultiOutputLinear,
    Linear,
    MultiOutputLinear,
    SegmentedLinear,
    TiedWeightLinear,
    TiedWeightLinearSequence,
    VectorizedLinear,
)
from tamm.layers.loss import FlattenedCrossEntropyLoss
from tamm.layers.norm import BatchNorm, L2Norm, LayerNorm, RMSNorm
from tamm.layers.pooler import (
    AdaptiveConvPooler,
    CAbstractorPooler,
    ConvPooler,
    SimpleAdaptiveAvgPooler,
)
from tamm.layers.positional_encoding import (
    AbsolutePositionalEmbedding,
    SpatialPositionalEmbedding,
)
from tamm.layers.residual import (
    GatedActivationResidualConnection,
    ResidualAdd,
    ShortcutAddActResidualConnection,
)
from tamm.layers.segmentation import (
    ConcatEmbeddingSegmentation,
    ConstantSegmentation,
    ConvEmbeddingPaddingTransform,
    TokensPaddingMask,
    UnionSegmentation,
)
from tamm.layers.sequential import Sequential
from tamm.layers.side_outputs import OutputWithSideOutputs
from tamm.layers.transformer import (
    ALiBiPositionalEncoding,
    AttentionMask,
    CausalLMTransformer,
    ExtendedTransformerStack,
    KVReuseTransformerLayerSequence,
    NoPositionalEncoding,
    RotaryPositionalEmbedding,
    SecondaryPositionalEncodings,
    SequentialPositionalEncoding,
    SinusoidalPositionalEmbedding,
    SlidingWindowPositionalEncoding,
    TextTransformerEncoder,
    TransformerAbsolutePositionalEmbedding,
    TransformerLayer,
    TransformerLayerSequence,
    TransformerPositionalEncoding,
    TransformerStack,
    TransformerTokenTypeEmbedding,
    TransformerTokenTypeEncoding,
    UniformTransformerLayerSequence,
    VisionTransformerEncoder,
    kv_cache,
)



import abc as _abc


class BaseLister(metaclass=_abc.ABCMeta):
    name = "unnamed"
    # pylint: disable=redefined-builtin

    def __init__(self, all, long, wide, show_deprecated):
        self.all = all
        self.long = long
        self.wide = wide
        self.show_deprecated = show_deprecated or all

    @_abc.abstractmethod
    def print(self):
        ...



"""
_adapters_v1.model_adapters.lora
================================

.. autoclass:: tamm._adapters_v1.model_adapters.lora.LoRAModelAdapter
    :show-inheritance:
    :members:
"""

# pylint: disable=no-member, cyclic-import

import logging as _logging
import warnings as _warnings
from collections import OrderedDict as _OrderedDict
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch.nn as _nn

from tamm._adapters_v1 import layer_annotations as _annotations
from tamm._adapters_v1.adapted_layer import AdaptedLayer as _AdaptedLayer
from tamm._adapters_v1.layer_adapters import BatchedLoRA as _BatchedLoRA
from tamm._adapters_v1.layer_adapters import (
    BatchedLoRAFusedMultiOutputLinear as _BatchedLoRAFusedMultiOutputLinear,
)
from tamm._adapters_v1.layer_adapters import LoRA as _LoRA
from tamm._adapters_v1.layer_adapters import (
    LoRAFusedMultiOutputLinear as _LoRAFusedMultiOutputLinear,
)
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.soft_mixing_lora import (
    SoftMixingLoRA as _SoftMixingLoRA,
)
from tamm._adapters_v1.layer_adapters.stack import (
    StackedOutputTransforms as _StackedOutputTransforms,
)
from tamm._adapters_v1.layer_annotations import LayerAnnotation as _LayerAnnotation
from tamm._adapters_v1.layer_annotations import (
    get_layer_annotations as _get_layer_annotations,
)
from tamm._adapters_v1.model_adapters.composition import (
    CompositeModelAdapterMixin as _CompositeModelAdapterMixin,
)
from tamm._adapters_v1.model_adapters.model_adapter import AdapterSpec as _AdapterSpec
from tamm._adapters_v1.model_adapters.model_adapter import ModelAdapter as _ModelAdapter

_logger = _logging.getLogger(__name__)


class LoRAModelAdapter(_ModelAdapter):
    """
    A model adapter which inserts :class:`_LoRA`
    and :class:`_LoRAFusedMultiOutputLinear` layers in a decoder only transformer model.
    """

    # pylint: disable=duplicate-code

    rank: int = 8
    alpha: _Optional[float] = None
    adapt_attention_queries: bool = True
    adapt_attention_keys: bool = True
    adapt_attention_values: bool = True
    adapt_attention_outputs: bool = True
    adapt_feed_forward_hidden_states: bool = True
    adapt_feed_forward_outputs: bool = True
    adapt_moe: bool = True
    dropout_p: float = 0
    layer_prefix: str = "layers.layer_"
    layer_pattern: _Optional[_Union[str, _List]] = None

    def _get_lora_linear_layer_type(
        self, layer_type: _List[_LayerAnnotation]
    ) -> _Tuple[_Optional[_LayerAnnotation], bool]:
        """
        Tests if the annotations match any of the supported layer annotation
        types for LoRA and if the config allows that layer type to be adapted.

        Returns the type of layer annotation and a boolean indicating
        if the layer is annotated with a batched layer annotation type.
        """
        masks = [
            self.adapt_attention_outputs,
            self.adapt_attention_queries,
            self.adapt_attention_keys,
            self.adapt_attention_values,
            self.adapt_feed_forward_hidden_states,
            self.adapt_feed_forward_outputs,
        ]
        annotation_types = [
            _annotations.AttentionOutputTransform,
            _annotations.AttentionInputTransformQuery,
            _annotations.AttentionInputTransformKey,
            _annotations.AttentionInputTransformValue,
            _annotations.FeedForwardHiddenTransform,
            _annotations.FeedForwardOutputTransform,
        ]
        batched_annotation_types = [
            _annotations.BatchedAttentionOutputTransform,
            _annotations.BatchedAttentionInputTransformQuery,
            _annotations.BatchedAttentionInputTransformKey,
            _annotations.BatchedAttentionInputTransformValue,
            _annotations.BatchedFeedForwardHiddenTransform,
            _annotations.BatchedFeedForwardOutputTransform,
        ]
        moe_annoation_types = [
            _annotations.MoEFeedForwardHiddenTransform,
            _annotations.MoEFeedForwardOutputTransform,
        ]
        for lt in layer_type:
            if any(
                (
                    mask and isinstance(lt, annotation_type)
                    for mask, annotation_type in zip(masks, annotation_types)
                )
            ):
                return lt, False
            if any(
                (
                    mask and isinstance(lt, annotation_type)
                    for mask, annotation_type in zip(masks, batched_annotation_types)
                )
            ):
                return lt, True
            if any(
                self.adapt_moe and isinstance(lt, annotation_type)
                for annotation_type in moe_annoation_types
            ):
                # MoE is implemented using BatchedLinear
                return lt, True
        return None, False

    def _get_lora_fused_multi_output_linear_type(
        self, layer_type: _List[_LayerAnnotation]
    ) -> _Tuple[_Optional[_LayerAnnotation], bool]:
        adapt_attention_input_transform = (
            self.adapt_attention_queries
            or self.adapt_attention_keys
            or self.adapt_attention_values
        )
        for lt in layer_type:
            if adapt_attention_input_transform and isinstance(
                lt, _annotations.AttentionInputTransformFused
            ):
                return lt, False
            if adapt_attention_input_transform and isinstance(
                lt, _annotations.BatchedAttentionInputTransformFused
            ):
                return lt, True
        return None, False

    def _create_adapters_impl(self, model: _nn.Module) -> _AdapterSpec:
        layer_adapters = {}
        for name, submodule in model.named_modules(remove_duplicate=True):
            if isinstance(submodule, _AdaptedLayer):
                if self.layer_pattern is not None:
                    layer_idx = (
                        int(name.split(self.layer_prefix)[-1].split(".")[0])
                        if self.layer_prefix in name
                        else None
                    )
                    if layer_idx is None:
                        _warnings.warn(
                            f"layer selector is specified but module's name: {name} "
                            f"does not match layer_prefix pattern: {self.layer_prefix}."
                            " Tamm will apply this adapter to all matched modules."
                        )
                    elif not self._adapt_current_layer(layer_idx):
                        continue

                adapter = self._create_adapter_for_submodule(submodule)
                if adapter is not None:
                    layer_adapters[name] = adapter
        return _AdapterSpec(layer_adapters=layer_adapters)

    def _adapt_current_layer(self, layer_idx: int):
        """
        Helper function to determine if we should
        adapt current module based on its layer_idx.
        """
        if isinstance(self.layer_pattern, str):
            if self.layer_pattern.lower() == "even":
                return layer_idx % 2 == 0
            if self.layer_pattern.lower() == "odd":
                return layer_idx % 2 != 0
            if ":" in self.layer_pattern:
                slice_values = self.layer_pattern.split(":")
                if len(slice_values) > 3:
                    raise RuntimeError(
                        "Found more than two `:` in layer_pattern. "
                        "Python slice only supports up to two `:`"
                    )
                start, end = int(slice_values[0]), int(slice_values[1])
                if len(slice_values) == 3:
                    return (
                        layer_idx
                        in list(range(end + 1))[slice(start, end, int(slice_values[2]))]
                    )
                if len(slice_values) == 2:
                    return layer_idx in list(range(end + 1))[slice(start, end)]

        if isinstance(self.layer_pattern, list):
            return layer_idx in self.layer_pattern

        raise ValueError(f"Unrecognized layer pattern {self.layer_pattern}")

    def _create_adapter_for_submodule(
        self, submodule: _nn.Module
    ) -> _Optional[_LayerAdapter]:
        wrapped_module = submodule.unwrap()
        layer_type = _get_layer_annotations(wrapped_module)
        linear_layer_type, is_batched = self._get_lora_linear_layer_type(layer_type)
        if linear_layer_type:
            input_dim = getattr(wrapped_module, linear_layer_type.input_dim_field)
            output_dim = getattr(wrapped_module, linear_layer_type.output_dim_field)
            if is_batched:
                vec_dim = getattr(wrapped_module, linear_layer_type.vec_dim_field)
                return _BatchedLoRA(
                    input_dim=input_dim,
                    output_dim=output_dim,
                    rank=self.rank,
                    vec_dim=vec_dim,
                    alpha=self.alpha,
                    dropout_p=self.dropout_p,
                    dtype=self.dtype,
                )
            return _LoRA(
                input_dim=input_dim,
                output_dim=output_dim,
                rank=self.rank,
                alpha=self.alpha,
                dropout_p=self.dropout_p,
                dtype=self.dtype,
            )
        (
            fused_linear_layer_type,
            is_batched,
        ) = self._get_lora_fused_multi_output_linear_type(layer_type)
        if fused_linear_layer_type:
            input_dim = getattr(wrapped_module, fused_linear_layer_type.input_dim_field)
            output_dims = getattr(
                wrapped_module, fused_linear_layer_type.output_dims_field
            )
            mask = (
                self.adapt_attention_queries,
                self.adapt_attention_keys,
                self.adapt_attention_values,
            )
            if is_batched:
                vec_dim = getattr(wrapped_module, fused_linear_layer_type.vec_dim_field)
                return _BatchedLoRAFusedMultiOutputLinear(
                    input_dim=input_dim,
                    output_dims=output_dims,
                    ranks=[self.rank] * 3,
                    vec_dim=vec_dim,
                    alphas=[self.alpha] * 3,
                    dropout_ps=[self.dropout_p] * 3,
                    mask=mask,
                    dtype=self.dtype,
                )
            return _LoRAFusedMultiOutputLinear(
                input_dim=input_dim,
                output_dims=output_dims,
                ranks=[self.rank] * 3,
                alphas=[self.alpha] * 3,
                dropout_ps=[self.dropout_p] * 3,
                mask=mask,
                dtype=self.dtype,
            )
        return None


class MultiLoRAModelAdapter(_CompositeModelAdapterMixin, _ModelAdapter):
    """
    Applies :py:class:`StackedOutputTransforms` composition to all modules which have
    LoRA adapters with ids specified in ``adapters``.
    The :py:class:`StackedOutputTransforms` adds up the results of
    multiple LoRA adapters in parallel to the adapted layer,
    such that the output of the adapted layer is:

    output = adapted_layer(input) + lora_0(inputs) +
        lora_1(inputs) + lora_2(inputs) + ...
    """

    adapters: _Dict[str, _ModelAdapter]

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _StackedOutputTransforms(_OrderedDict(child_adapters))


class SoftMixingLoRAModelAdapter(_CompositeModelAdapterMixin, _ModelAdapter):
    """
    Applies :py:class:`SoftMixingLoRA` composition to all modules which have
    adapters with ids specified in ``adapters``. The :py:class:`SoftMixingLoRA`
    applies a mixture of multiple LoRA adapters in parallel to the
    adapted layer, such that the output of the adapted layer is:

    output = adapted_layer(input) + lora_0(weight[:, 0] * inputs) +
        lora_1(weight[:, 1] * inputs) + lora_2(weight[:, 2] * inputs) + ...

    where weight is a m x n tensor where m is equal to the batch size of inputs
    and n is equal to the number of LoRA adapters being applied together.
    """

    adapters: _Dict[str, _ModelAdapter]

    def _create_composite_adapter(self, child_adapters: _Dict[str, _LayerAdapter]):
        return _SoftMixingLoRA(_OrderedDict(child_adapters))



"""
Modeling Utilities
------------------

.. autoclass:: tamm.models.common.ModelConfig
    :members: create_builder, create_model

.. autoclass:: tamm.models.common.ModelMetadata
    :members:
    :show-inheritance:

.. autoclass:: tamm.models.common.ModelMixin
    :members:

"""
from tamm.models.common.config import ModelConfig
from tamm.models.common.metadata import ModelMetadata
from tamm.models.common.mixin import ModelMixin

__all__ = ["ModelMixin", "ModelMetadata", "ModelConfig"]



"""
.. _adapted_layer:

adapters.adapted_layer
======================

.. autoclass:: tamm.adapters.AdaptedLayer
    :show-inheritance:
    :members:
"""

import logging as _logging
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch.nn as _nn

from tamm._adapters_v1.layer_adapters.common import AdapterMode as _AdapterMode
from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.common import (
    MergeableLayerAdapterMixin as _MergeableLayerAdapterMixin,
)

_logger = _logging.getLogger(__name__)


class AdaptedLayer(_nn.Module):
    """
    A wrapper layer for enabling adapters for a module. Provides functionality
    for managing adapters and modifies either the inputs or the outputs (or both) of
    the wrapped model using the currently active adapter.
    """

    def __init__(self, layer: _nn.Module):
        super().__init__()
        self.wrapped = layer
        self.adapters = _nn.ModuleDict()
        self._active_adapter_id = None

    @staticmethod
    def _get_parent_child_adapter_id(adapter_id: str) -> _Optional[_Tuple[str, str]]:
        children = adapter_id.split(".")
        if len(children) > 1:
            return children[0], ".".join(children[1:])
        return None

    def unwrap(self) -> _nn.Module:
        """
        Returns the layer wrapped in this :py:class:`AdaptedLayer`
        """
        return self.wrapped

    def add_adapter(
        self,
        adapter_id: str,
        adapter: _LayerAdapter,
    ):
        """
        Add an adapter to the layer.

        Args:
            adapter_id (:obj:`str`): A unique identifier for the adapter being added.
            adapter: (:py:class:`_LayerAdapter` ): Adapter to be added to
              this layer.
        """
        self.adapters[adapter_id] = adapter

    def _get_adapter_impl(self, adapter_id: str) -> _Optional[_LayerAdapter]:
        parent_child_id = self._get_parent_child_adapter_id(adapter_id)
        if parent_child_id:
            parent_id, child_id = parent_child_id
            if parent_id not in self.adapters:
                return None
            try:
                return self.adapters[parent_id].get_submodule(child_id)
            except AttributeError:
                return None
        if adapter_id in self.adapters:
            return self.adapters[adapter_id]
        return None

    def _is_nested_adapter_id(self, adapter_id: str) -> bool:
        """
        Returns True if the adapter id is a nested id, such as
        adapter_a.child_1.child2
        """
        return self._get_parent_child_adapter_id(adapter_id) is not None

    def has_adapter(self, adapter_id: str) -> bool:
        """
        Returns True if an adapter identified by the ``adapter_id`` exists.
        Args:
            adapter_id (:obj:`str`): The unique identifier for the desired adapter.
        """

        return self._get_adapter_impl(adapter_id) is not None

    def get_adapter(self, adapter_id: str) -> _LayerAdapter:
        """
        Returns the adapter identified by the ``adapter_id``, if it exists.
        Args:
            adapter_id (:obj:`str`): The unique identifier for the desired adapter.
        """
        adapter = self._get_adapter_impl(adapter_id)
        if not adapter:
            raise AttributeError(
                f"Adapter with id: {adapter_id} does not exist in the adapted layer."
            )
        return adapter

    def delete_adapter(self, adapter_id: str):
        """
        Deletes an adapter identified by ``adapter_id``, if it exists.

        Args:
            adapter_id (:obj:`str`): The unique identifier for the adapter being
              deleted.
        """
        if self._is_nested_adapter_id(adapter_id):
            raise AttributeError(
                f"Received nested adapter id: {adapter_id}. delete_adapter only"
                f"supports un-nested adapter_ids with no '.' in the id."
            )
        self.get_adapter(adapter_id)
        self.adapters.pop(adapter_id)
        if self._active_adapter_id == adapter_id:
            self._active_adapter_id = None

    def freeze_adapter(self, adapter_id: str):
        """
        Freeze the parameters of the adapter identified by ``adapter_id``.

        Args:
            adapter_id (:obj:`str`): The unique identifier for the adapter being frozen.
        """
        self.get_adapter(adapter_id).freeze()

    def unfreeze_adapter(self, adapter_id: str):
        """
        Make the parameters of the adapter identified by ``adapter_id`` trainable.

        Args:
            adapter_id (:obj:`str`): The unique identifier for the adapter
              being made trainable.
        """
        self.get_adapter(adapter_id).unfreeze()

    @property
    def active_adapter(self):
        """
        Returns the currently active adapter.
        """
        if self._active_adapter_id is None:
            return None
        return self.get_adapter(self._active_adapter_id)

    @property
    def active_adapter_id(self):
        """
        Returns the id of the currently active adapter.
        """
        return self._active_adapter_id

    def set_active_adapter(self, adapter_id: str, replace: bool = True):
        """
        Set one of the adapters already added to the layer as the active adapter.
        All other adapters besides this are inactive and won't be used during
        model's forward pass.

        Args:
            adapter_id (:obj:`str`): The unique identifier for the adapter
              to be activated.
            replace (:obj:`bool`): When set to ``True``, the current active adapter is replaced
              by the adapter with ``adapter_id``. Throws an exception when ``False`` if an
              active adapter already exists.
        """
        # call get adapter to make sure the adapter exists
        self.get_adapter(adapter_id)
        if self._is_nested_adapter_id(adapter_id):
            raise AttributeError(
                f"Received nested adapter id: {adapter_id}. set_active_adapter only"
                f"supports un-nested adapter_ids with no '.' in the id."
            )
        if replace:
            self.unset_active_adapter()
        elif (
            not replace
            and self._active_adapter_id is not None
            and adapter_id != self._active_adapter_id
        ):
            raise ValueError(
                f"Found an existing active adapter {self._active_adapter_id}. For replacing "
                f"the existing active adapter, call set_active_adapter with replace=True. "
                "Only one adapter can be activated at a time. For combining multiple adapters, "
                "consider using composite adapters, such as CompositeInputTransform or CompositeInputTransform."
            )
        self._active_adapter_id = adapter_id

    def unset_active_adapter(self):
        """
        Unset the active adapter from the layer, if any.
        """
        if self._active_adapter_id is not None:
            _logger.debug(f"De-activating adapter with id: {self._active_adapter_id}.")
            self._active_adapter_id = None

    def _merge_adapter(self, adapter_id: str):
        """
        Merge weights of adapter with adapter_id into the wrapped module,
        if the adapter supports merging.

         Args:
            adapter_id (:obj:`str`): The unique identifier for the adapter
              to be merged.
        """
        if self._is_nested_adapter_id(adapter_id):
            raise AttributeError(
                f"Received nested adapter id: {adapter_id}. merge_adapter only"
                f"supports un-nested adapter_ids with no '.' in the id."
            )
        adapter = self.get_adapter(adapter_id)
        if not isinstance(adapter, _MergeableLayerAdapterMixin):
            raise ValueError(  # pylint: disable=raise-missing-from
                f"Adapter with id: {adapter_id} of type: {type(adapter)}"
                f" doesn't support merge operation."
            )
        adapter.merge_adapter(self.wrapped)
        self.delete_adapter(adapter_id)

    def forward(self, *args, **kwargs):
        """
        Modifies either the inputs or the outputs (or both) of the wrapped
        model with the currently active adapter.
        """
        if self.active_adapter is None:
            return self.wrapped(*args, **kwargs)

        # pylint: disable=not-callable
        if self.active_adapter.has_input_transform:
            transformed_args, transformed_kwargs = self.active_adapter(
                mode=_AdapterMode.TRANSFORM_INPUTS,
                args=args,
                kwargs=kwargs,
                transformed_args=None,
                transformed_kwargs=None,
                outputs=None,
            )
        else:
            # The base input transform is a noop, but we avoid calling it
            # because if active_adapter is wrapped with FSDP v1, it triggers
            # an extra parameter unshard (which can slow performance)
            transformed_args, transformed_kwargs = args, kwargs

        model_out = self.wrapped(*transformed_args, **transformed_kwargs)

        if self.active_adapter.has_output_transform:
            return self.active_adapter(
                mode=_AdapterMode.TRANSFORM_OUTPUTS,
                args=args,
                kwargs=kwargs,
                transformed_args=transformed_args,
                transformed_kwargs=transformed_kwargs,
                outputs=model_out,
            )
        return model_out



"""
Layer Utilities
---------------

.. autoclass:: tamm.layers.common.LayerBuilder
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.LayerConfig
    :members: create_builder, create_layer

.. autoclass:: tamm.layers.common.LayerMixin
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.BuildableMixin
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.common.ConfigurableLayerMixin
    :members:

.. autoclass:: tamm.layers.common.PretrainedLoader


Module Markers
^^^^^^^^^^^^^^

.. autofunction:: tamm.layers.common.init_marker
.. autofunction:: tamm.layers.common.update_marker
.. autofunction:: tamm.layers.common.get_marker

.. autoclass:: tamm.layers.common.ModuleMarker
    :members:

.. automodule:: tamm.layers.common.context_hooks
.. automodule:: tamm.layers.common.post_hooks
"""
from tamm.layers.common._marker import (
    ModuleMarker,
    get_marker,
    init_marker,
    update_marker,
)
from tamm.layers.common.builder import BuildableMixin, LayerBuilder
from tamm.layers.common.config import LayerConfig, _BaseConfig
from tamm.layers.common.context_hooks import (
    DefaultDeviceContextHook,
    DtypeContextHook,
    PretrainedContextHook,
    UseMetaInitTrickContextHook,
)
from tamm.layers.common.mixins import (
    ConfigurableLayerMixin,
    LayerMixin,
    _BaseConfigurableMixin,
)
from tamm.layers.common.post_hooks import (
    ArchOptimizersPostHook,
    AttachConfigPostHook,
    AttachMetadataPostHook,
    FreezeParamsPostHook,
    ModelInitializerPostHook,
    get_model_adapters_post_hook,
)
from tamm.layers.common.pretrained_loader import PretrainedLoader

__all__ = [
    "LayerBuilder",
    "LayerConfig",
    "LayerBuilder",
    "LayerMixin",
    "BuildableMixin",
    "ConfigurableLayerMixin",
    "ArchOptimizersPostHook",
    "AttachConfigPostHook",
    "AttachMetadataPostHook",
    "FreezeParamsPostHook",
    "ModelInitializerPostHook",
    "get_model_adapters_post_hook",
    "DefaultDeviceContextHook",
    "PretrainedContextHook",
    "UseMetaInitTrickContextHook",
    "PretrainedLoader",
    "DtypeContextHook",
    "ModuleMarker",
    "init_marker",
    "update_marker",
    "get_marker",
    "_BaseConfig",
    "_BaseConfigurableMixin",
]



"""
Model Repository
================

.. automodule:: tamm.model_repo.publishing

Backends
--------

.. autoclass:: tamm.model_repo.RegistryModelRepo
    :members:
    :show-inheritance:

Base Class
----------
.. automodule:: tamm.model_repo.model_repo
    :members:

API
---

.. autofunction:: tamm.model_repo.get_model_repo_lazy
.. autofunction:: tamm.model_repo.get_model_repo_eager

"""
from tamm.model_repo.api import (
    create_adapted_model,
    create_model,
    create_model_builder,
    create_model_config,
    get_model_repo_eager,
    get_model_repo_lazy,
    is_adapted_model_name,
    is_model_builder_name,
    is_model_config_name,
    is_model_name,
    list_adapted_models,
    list_available_lora_adapted_models,
    list_model_builders,
    list_model_configs,
    list_models,
)
from tamm.model_repo.model_repo import CompositeModelRepo
from tamm.model_repo.publishing import PublishedModelConfig
from tamm.model_repo.registry import RegistryModelRepo

__all__ = (
    "create_model",
    "create_model_config",
    "create_adapted_model",
    "create_model_builder",
    "list_model_configs",
    "is_model_config_name",
    "is_model_name",
    "is_adapted_model_name",
    "is_model_builder_name",
    "list_adapted_models",
    "list_available_lora_adapted_models",
    "list_model_builders",
    "list_models",
    "get_model_repo_lazy",
    "get_model_repo_eager",
    "RegistryModelRepo",
    "PublishedModelConfig",
    "CompositeModelRepo",
)



"""
AO Layers
=========

.. autoclass:: tamm.ao.layers.KVQuantizer
    :members:

.. autoclass:: tamm.ao.layers.QuantizingKVCacher
    :members:

.. autoclass:: tamm.ao.layers.FakeQuantize
    :members:

.. autoclass:: tamm.ao.layers.SimpleEMAMinMaxObserver
    :members:

.. autoclass:: tamm.ao.layers.SimpleEMAPerChannelMinMaxObserver
    :members:

.. autofunction:: tamm.ao.layers.functional.fake_quantize
"""

import tamm.ao.layers.functional
from tamm.ao.layers.fake_quantize import (
    FakeQuantize,
    SimpleEMAMinMaxObserver,
    SimpleEMAPerChannelMinMaxObserver,
)
from tamm.ao.layers.kv_quantizer import KVQuantizer
from tamm.ao.layers.quantizing_kv_cacher import QuantizingKVCacher

__all__ = [
    "KVQuantizer",
    "QuantizingKVCacher",
    "FakeQuantize",
    "SimpleEMAMinMaxObserver",
    "SimpleEMAPerChannelMinMaxObserver",
]



import dataclasses as _dataclasses
import logging as _logging
import warnings as _warnings

import torch.export as _torch_export

_logger = _logging.getLogger(__name__)


def torch_exportable_dataclass(cls):
    """
    A decorator that applies both ``@dataclasses.dataclass`` and
    ``@torch.export.register_dataclass``.  This is helpful for making dataclasses
    that are compatible with ``torch.export``.
    """

    cls = _dataclasses.dataclass(cls)

    with _warnings.catch_warnings():
        _warnings.simplefilter("ignore")

        try:
            _torch_export.register_dataclass(cls)
        except AttributeError:
            # register_dataclass not available in some torch versions
            try:
                # pylint: disable=import-outside-toplevel,protected-access
                import torch._export.utils

                torch._export.utils.register_dataclass_as_pytree_node(cls)
            except Exception:
                _logger.debug("Could not register dataclass %s with torch.export", cls)

    return cls



import dataclasses as _dataclasses
import logging as _logging
from typing import TYPE_CHECKING
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import _helpers
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin

if TYPE_CHECKING:
    from tamm.preprocessors import Preprocessor
    from tamm.tokenizers.common import TokenizerConfig

_logger = _logging.getLogger(__name__)


def maybe_supersede_tokenizer_spec(new_cls: "ModelMetadata") -> "ModelMetadata":
    """
    Attempt to overwrite .tokenizer_spec with .preprocessor_spec.tokenizer_spec if
    `.preprocessor_spec.tokenizer_spec` is properly defined.

    Args:
        new_cls: A newly created :class:`ModelMetadata`

    Returns: :class:`ModelMetadata`

    """
    try:
        tokenizer_spec = new_cls.preprocessor_spec.tokenizer_spec
    except AttributeError:
        return new_cls
    if new_cls.tokenizer_spec is None:
        return _dataclasses.replace(new_cls, tokenizer_spec=tokenizer_spec)
    if tokenizer_spec == new_cls.tokenizer_spec:
        return new_cls

    _logger.warning(
        "`.preprocessor_spec.tokenizer_spec` supersedes `.tokenize_spec`"
        " for this ModelMetadata. "
        "Ask model owner to address the discrepancy in model config. "
        "Set debug mode to see details."
    )
    _logger.debug(
        "model_metadata = %s",
        new_cls,
    )
    _logger.debug(
        "`.preprocessor_spec.tokenizer_spec` = %s",
        new_cls.preprocessor_spec.tokenizer_spec,
    )
    _logger.debug("`.tokenizer_spec` = %s", new_cls.tokenizer_spec)
    return _dataclasses.replace(new_cls, tokenizer_spec=tokenizer_spec)


@_dataclasses.dataclass(frozen=True, repr=False)
class ModelMetadata(
    _helpers.DataClassReprMixin,
    _JSONSerializableMixin,
    json_namespace="models",
):
    tokenizer_spec: _Optional[_Union[str, "TokenizerConfig"]] = None
    preprocessor_spec: _Optional[
        _Union[_Dict[str, "Preprocessor"], str, "Preprocessor"]
    ] = None
    source_model_details: _Dict[str, _Any] = _dataclasses.field(default_factory=dict)
    source_model_tamm_id: _Optional[str] = None

    def _to_json_dict_impl(self):
        result = _helpers.dataclass_to_dict(self, omit_defaults=True)
        if len(self.source_model_details) == 0:
            result.pop("source_model_details")
        return result

    @classmethod
    def _from_json_dict_impl(cls, **raw_dict: dict):
        new_cls = _helpers.dataclass_init_drop_missing_keys(
            raw_dict, dataclass_type=cls
        )
        new_cls = maybe_supersede_tokenizer_spec(new_cls)
        return new_cls

    @property
    def is_empty(self):
        return (
            len(self.source_model_details) == 0
            and self.tokenizer_spec is None
            and self.source_model_tamm_id is None
        )



"""
Mixins for Hugging Face transformers wrappers.
"""

import logging as _logging
import os as _os
from typing import Optional as _Optional

import transformers as _transformers

from tamm import _helpers, model_repo
from tamm.hf.transformers.models import registry as _registry
from tamm.runtime_configuration import rc as _rc
from tamm.typing import PathLike as _PathLike

_logger = _logging.getLogger(__name__)


class FromModelIDMixin:
    """
    A mixin for adding :meth:`from_tamm_model_id` and
    :meth:`save_pretrained_for_tamm_model_id` methods to
    :class:`transformers.PreTrainedModel` subclasses.
    """

    @classmethod
    def from_tamm_model_id(
        cls, model_id: str, *model_args, **kwargs
    ) -> _transformers.PreTrainedModel:
        """
        Creates a Hugging Face version of a tamm model.  This method fetches pretrained
        weights and stores them locally in ``$HOME/.tamm`` to speed up future calls. It
        supports standard Hugging Face options for :meth:`from_pretrained`.

        Args:
            model_id (:obj:`str`): The tamm model id.
            *model_args, **kwargs: Arguments to pass to
                :meth:`transformers.PreTrainedModel.from_pretrained` when initializing
                the model.

        Returns:
            A :obj:`transformers.PreTrainedModel` instance of the tamm model.
        """
        save_directory = cls.save_pretrained_for_tamm_model_id(model_id)
        return cls.from_pretrained(save_directory, *model_args, **kwargs)

    @classmethod
    def save_pretrained_for_tamm_model_id(
        cls, model_id: str, save_directory: _Optional[_PathLike] = None
    ) -> _PathLike:
        """
        Saves a "pretrained" directory for loading tamm models with Hugging Face
        transformers :meth:`from_pretrained` methods.

        Args:
            model_id (:obj:`str`): The tamm model id.
            save_directory (:obj:`PathLike`, optional): The path for saving the
                pretrained model.  If ``None``, the method uses a default path in
                ``$HOME/.tamm``.  If the directory already exists, this function assumes
                it is a cached copy of the model and returns without saving a new one.

        Returns:
            The path to the pretrained model.
        """

        if save_directory is None:
            save_directory = _os.path.join(_rc.user_dir, "hf", model_id)
        lock_path = _os.path.join(_rc.user_dir, "hf", f"{model_id}.cache_lock")

        with _helpers.file_lock(lock_path):
            if _os.path.exists(save_directory):
                _logger.info("Found cached model at %s", save_directory)
                return save_directory

            config = model_repo.create_model_config(model_id)
            # pylint: disable-next=protected-access
            hf_cls = _registry._get_hf_model_type(config.__class__)
            model = hf_cls.from_tamm_config(config)

            _logger.info("Writing model to %s", save_directory)
            model.save_pretrained(save_directory)
            return save_directory



from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import kv_cache as _kv_cache
from tamm.layers.transformer.token_metadata import common as _common
from tamm.typing import OptionalTensor as _OptionalTensor


class VanillaTokenMetadataLogic(_common.TokenMetadataLogic, _LayerMixin):
    """
    This :class:`TokenMetadataLogic` subclass defines the default positions
    as the number of times a token's segment ID has previously occurred in
    ``segment_ids``.  The layer also reads and writes segment ids from the
    ``kv_cache``.
    """

    def _from_q_values(
        self,
        segment_ids: _OptionalTensor = None,
        default_segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        aux_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> _common.TokenMetadata:
        """
        Factory method for creating segments and positions without a KV cache.  In this
        case q and kv tensors have the same length.  If positions is None, this defaults
        to 0, 1, 2, ... for each segment of each sequence.
        """
        if positions is None:
            if segment_ids is None:
                batch_size, q_length = default_segment_ids.shape
                positions = (
                    _torch.arange(
                        q_length, device=default_segment_ids.device, dtype=_torch.int64
                    )
                    .unsqueeze(0)
                    .expand(batch_size, q_length)
                )
            else:
                positions = self._compute_default_q_positions(
                    q_segment_ids=segment_ids, kv_segment_ids=segment_ids
                )
        if segment_ids is None:
            segment_ids = default_segment_ids
        return _common.TokenMetadata(
            q_segment_ids=segment_ids,
            kv_segment_ids=segment_ids,
            q_positions=positions,
            kv_positions=positions,
            q_token_types=token_types,
            kv_token_types=token_types,
            aux_metadata=aux_metadata,
        )

    def _from_kv_cache(
        self,
        kv_cache: _Optional[_kv_cache.BaseKVCacheView],
        *,
        segment_ids: _OptionalTensor = None,
        default_segment_ids: _OptionalTensor = None,
        positions: _OptionalTensor = None,
        token_types: _OptionalTensor = None,
        aux_metadata: _Optional[_Dict[str, _Any]] = None,
    ) -> _common.TokenMetadata:
        """
        Factory method for creating segments and positions with a KV cache.  In this
        case kv tensors may have a longer length than k tensors.
        """
        if segment_ids is None:
            segment_ids = default_segment_ids

        q_length = segment_ids.size(-1)

        kv_cache.write_segment_ids_to_tail(segment_ids)
        kv_segment_ids = kv_cache.read_segment_ids()

        if positions is None:
            positions = self._compute_default_q_positions(
                q_segment_ids=segment_ids, kv_segment_ids=kv_segment_ids
            )
        kv_cache.write_positions_to_tail(positions)
        kv_positions = kv_cache.read_positions()

        if token_types is not None:
            kv_cache.write_token_types_to_tail(token_types)
            kv_token_types = kv_cache.read_token_types()
            q_token_types = kv_token_types[:, -q_length:]
        else:
            q_token_types = kv_token_types = None

        return _common.TokenMetadata(
            q_segment_ids=kv_segment_ids[..., -q_length:],
            kv_segment_ids=kv_segment_ids,
            q_positions=kv_positions[..., -q_length:],
            kv_positions=kv_positions,
            q_token_types=q_token_types,
            kv_token_types=kv_token_types,
            aux_metadata=aux_metadata,
        )

    # pylint: disable=too-many-locals
    @staticmethod
    def _compute_default_q_positions(
        *, q_segment_ids: _torch.Tensor, kv_segment_ids: _torch.Tensor
    ):
        """
        Computes default positions corresponding to q_segment_ids.  The tail values of
        kv_segment_ids should be the same as q_segment_ids.  Each position value of the
        result equals the number of times the corresponding segment ID has previously
        occurred in kv_segment_ids.

        Example:

            q_segment_ids: [[2, 1, 1, 0]]
            kv_segment_ids: [[0, 0, 0, 0, 0, 1, 1, 2, 1, 1, 0]]
            result: [[0, 2, 3, 5]]
        """
        device = q_segment_ids.device

        if (
            kv_segment_ids.device.type in ("cuda", "mps")
            and kv_segment_ids.dtype is _torch.bool
        ):
            kv_segment_ids = kv_segment_ids.type(
                _torch.uint8
                # older torch versions (2.1) do not support sort with cuda and bool
            )

        sorted_kv_segment_ids, order = kv_segment_ids.sort(dim=-1, stable=True)
        sorted_indices = _torch.arange(kv_segment_ids.size(-1), device=device)

        # find indices where segment id changes in the sorted segment ids
        if device.type == "mps":
            # mps does not support cummax as of torch 2.3
            segment_id_change_idx = _torch.searchsorted(
                sorted_kv_segment_ids, sorted_kv_segment_ids, right=False
            )
        else:
            is_different_segment_id = (
                sorted_kv_segment_ids != sorted_kv_segment_ids.roll(shifts=1, dims=-1)
            )
            segment_id_change_idx = sorted_indices * is_different_segment_id
            segment_id_change_idx, _ = segment_id_change_idx.cummax(dim=-1)

        # subtract the change index from sorted positions to make sure each segment id group begins from 0
        positions_sorted = sorted_indices - segment_id_change_idx
        result = positions_sorted.scatter(-1, order, positions_sorted)
        return result[..., -q_segment_ids.size(-1) :]



import logging as _logging
from os import PathLike as _PathLike
from typing import Union as _Union

import torch as _torch

from tamm import _helpers
from tamm.converters.models import common as _common
from tamm.converters.models import registry as _registry
from tamm.typing import OptionalDtypeOrString as _OptionalDtypeOrString
from tamm.typing import StateDictType as _StateDictType

_logger = _logging.getLogger(__name__)


def convert_from_tamm_state_dict(
    state_dict: _StateDictType, converter: _Union[str, _common.ModelStateDictConverter]
) -> _StateDictType:
    """
    Converts a tamm state dict to another format.

    Args:
        state_dict (:obj:`StateDictType`): A state dict for a tamm model.
        converter (:obj:`str` or :obj:`.ModelStateDictConverter`, optional): A converter
            id or converter instance for converting ``obj`` to a different format.

    Returns:
        :obj:``StateDictType``: The converted state dict.
    """
    if not isinstance(converter, _common.ModelStateDictConverter):
        converter = _registry.create_converter_from_tamm_state_dict(
            converter, state_dict
        )
    _logger.info(
        f"Converting tamm state dict to new format with {converter.__class__.__name__}."
    )
    return converter.convert_from_tamm(state_dict)


def save(
    state_dict: _StateDictType,
    f: _Union[str, _PathLike],
    *,
    converter: _Union[str, _common.ModelStateDictConverter] = None,
    dtype: _OptionalDtypeOrString = None,
    **kwargs,
):
    """
    Saves a tamm state dict to a converted format on disk.

    Args:
        state_dict (:obj:`StateDictType`): A state dict for a tamm model, which maps
            parameter names to parameter values.
        f (:obj:`str` or :obj:`PathLike`): The filepath for saving the converted state.
        converter (:obj:`str` or :obj:`.ModelStateDictConverter`, optional): A converter
            id or converter instance for converting ``obj`` to a non-tamm format.  This
            value is ``None`` by default, in which case this function behaves like
            :func:`torch.save` .  Use :func:`tamm.list_converters` or
            ``tamm ls converters -l`` for a list of available converter ids.
        dtype (:obj:`torch.dtype` or :obj:`str`, optional): Output dtype, such as
            ``float16``, ``float32``, etc.  Before converting ``state_dict`` to the
            other format, the function casts the floating point tensors to this dtype.
            The default is ``None``, which results in no cast.
        **kwargs: Keyword arguments to forward to the :meth:`save_other_state_dict`
            method of the converter.
    """
    _helpers.maybe_cast_state_dict_(state_dict, dtype=dtype)

    if converter is None:
        _torch.save(state_dict, f, **kwargs)
        _logger.info(f"Saved tamm state dict to {f} without conversion.")
        return

    if not isinstance(converter, _common.ModelStateDictConverter):
        converter = _registry.create_converter_from_tamm_state_dict(
            converter, state_dict
        )
    state_dict = convert_from_tamm_state_dict(state_dict, converter=converter)
    _logger.info(f"Saving converted state to {f}.")
    converter.save_other_state_dict(state_dict, f, **kwargs)



from typing import Dict as _Dict
from typing import Optional as _Optional

import torch as _torch

from tamm import _helpers
from tamm.layers import basic as _basic
from tamm.layers import linear as _linear
from tamm.layers import torch_nn as _torch_nn
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.layers.mixture_of_experts import loss as _loss
from tamm.layers.mixture_of_experts import sampler as _sampler
from tamm.utils import torch_utils as _torch_utils


@_torch_utils.torch_exportable_dataclass
class MixtureOfExpertsRouterOutput:
    """
    Result dataclass for :obj:`MixtureOfExpertsRouter` layers.

    Args:
        expert_assignments (:obj:`torch.Tensor`): An integer tensor that
            maps tokens to chosen experts.  The expected shape is
            ``(*batch_shape, seq_len, num_experts_per_token)``.
        assignment_weights (:obj:`torch.Tensor`): A floating point tensor with
            the same shape as ``expert_assignments``.  These are weights
            for combining expert outputs for each token.
        losses (:obj:`torch.Tensor`): A dictionary of scalar auxiliary
            losses for training router behavior.
    """

    expert_assignments: _torch.Tensor
    assignment_weights: _torch.Tensor
    losses: _Dict[str, _torch.Tensor]


class MixtureOfExpertsRouter(_torch.nn.Module, _ConfigurableLayerMixin):
    """
    A layer that computes expert assignments for each token. This layer is a
    sequence of the following child layers:

    * ``input_transform``
    * ``logits_cap`` (optional)
    * ``softmax``
    * ``sampler``

    It also computes logits and load balancing losses.
    """

    def __init__(
        self,
        *,
        input_transform,
        logits_cap=None,
        softmax,
        sampler,
        logits_loss=None,
        load_balance_loss=None,
        cast_dtype=_torch.float32,
    ):
        super().__init__()
        _helpers.append_children(
            self,
            input_transform=input_transform,
            logits_cap=logits_cap,
            softmax=softmax,
            sampler=sampler,
            logits_loss=logits_loss,
            load_balance_loss=load_balance_loss,
        )
        self.cast_dtype = cast_dtype

    def extra_repr(self):
        return f"cast_dtype={self.cast_dtype}"

    def forward(self, tensor: _torch.Tensor) -> MixtureOfExpertsRouterOutput:
        """
        Transforms hidden states into per-token expert assignments and weights.

        Args:
            tensor (:obj:`torch.Tensor`): Input hidden states with shape
                ``(*batch_shape, seq_len, hidden_dim)``.

        Returns:
            A :obj:`MixtureOfExpertsRouterOutput`.
        """

        losses = {}

        logits = self.input_transform(tensor)

        if self.cast_dtype is not None:
            logits = logits.type(self.cast_dtype)

        if self.logits_cap is not None:
            logits = self.logits_cap(logits)

        if self.logits_loss is not None:
            losses["logits_loss"] = self.logits_loss(logits)

        probabilities = self.softmax(logits)

        sampler_output = self.sampler(probabilities)

        if self.load_balance_loss is not None:
            losses["load_balance_loss"] = self.load_balance_loss(
                expert_assignments=sampler_output.expert_assignments,
                assignment_probabilities=probabilities,
            )

        return MixtureOfExpertsRouterOutput(
            expert_assignments=sampler_output.expert_assignments,
            assignment_weights=sampler_output.assignment_weights,
            losses=losses,
        )

    @classmethod
    def create_basic_builder(  # pylint: disable=arguments-differ
        cls,
        *,
        input_dim: int,
        num_experts: int,
        num_experts_per_token: int,
        logits_cap: _Optional[float] = None,
        include_logits_loss: bool = True,
        include_load_balance_loss: bool = True,
    ):
        """
        Configures a basic MoE router with top-k sampling.

        Args:
            input_dim (:obj:`int`): The hidden dimension of the input states.
            num_experts (:obj:`int`): The number of experts.
            num_experts_per_token (:obj:`int`): The number of experts to
                activate for each token.
            logits_cap (:obj:`float`, optional): An optional cap value for
                the magnitude of router logits.  If specified, the layer
                applies tanh soft-capping to the logits.  Defaults to
                ``None`` (no capping).
            include_logits_loss (:obj:`bool`, optional): A flag for including
                a :obj:`RouterZLoss` value in the layer's outputs.  Defaults
                to ``True``.
            include_load_balance_loss (:obj:`bool`, optional): A flag for
                including a :obj:`LoadBalanceLoss` value in the layer's
                outputs.  Defaults to ``True``.
        """

        kwargs = {}

        kwargs["input_transform"] = _linear.Linear.Builder(
            input_dim, num_experts, bias=False
        )

        if logits_cap is not None:
            kwargs["logits_cap"] = _basic.SoftCap.Builder(cap=logits_cap)

        kwargs["softmax"] = _torch_nn.Softmax.Builder(dim=-1)

        kwargs["sampler"] = _sampler.TopKMixtureOfExpertsSampler.Builder(
            num_experts_per_token=num_experts_per_token
        )

        if include_logits_loss:
            kwargs["logits_loss"] = _loss.RouterZLoss.Builder()

        if include_load_balance_loss:
            kwargs["load_balance_loss"] = _loss.LoadBalanceLoss.Builder()

        return cls.Builder(**kwargs)



"""
This module provides classes that help with decoding, such as a KVCache handler
"""

import logging as _logging
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch
import torch.nn as _nn

from tamm.layers.common import LayerMixin as _LayerMixin
from tamm.layers.transformer import kv_cache as _kv_cache

_logger = _logging.getLogger(__name__)


class KVCacher(_nn.Module, _LayerMixin):
    """
    Reads from and updates a KV cache.

    The KVCache is expected to be a tensor with shape 2 x batch x S x dim,
    where S is > than the sequence dimension of query/key/value.

    The values in the new key and value tensors are stored in the tail of the
    cache, then the full cache is returned.
    """

    def __init__(self):
        super().__init__()

    def forward(
        self,
        query: _torch.Tensor,
        key: _torch.Tensor,
        value: _torch.Tensor,
        kv_cache: _Optional[_kv_cache.KVCacheLayerView] = None,
    ) -> _Tuple[_torch.Tensor, _torch.Tensor, _torch.Tensor]:
        """If kv_cache is not None, write key and value to the tail of the cache,
        and return the full cache for key and value.

        Args:
            query (:obj:`torch.Tensor`): batch x L x dim
            key (:obj:`torch.Tensor`): batch x L x dim
            value (:obj:`torch.Tensor`): batch x L x dim
            kv_cache (:obj: `Optional[torch.Tensor]`): 2 x batch x S x dim, where S >= L

        Returns:
            if kv_cache is none, Tuple of 3 tensors all batch x L x dim
            else Tuple of tensors [batch x L x dim, batch x S x dim, batch x S x dim]

        """
        if kv_cache is not None:
            kv_cache.write_to_tail(key, value)
            return query, *kv_cache.read()
        return query, key, value



import inspect as _inspect
import logging as _logging
from typing import TYPE_CHECKING
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

from tamm.layers.common import _BaseConfigurableMixin
from tamm.layers.common import post_hooks as _post_hooks
from tamm.models.common.config import ModelConfig as _ModelConfig
from tamm.models.common.metadata import ModelMetadata as _ModelMetadata
from tamm.utils import OptionalBool as _OptionalBool

if TYPE_CHECKING:
    from tamm._adapters_v1 import ModelAdapter
    from tamm.ao import ArchOptimizer
    from tamm.typing import (
        LenientOptionalBool,
        OptionalDeviceOrString,
        OptionalDtypeOrString,
    )

_logger = _logging.getLogger(__name__)


class ModelMixin(_BaseConfigurableMixin):
    """
    A mixin for common functionality of |tamm| models.  Model subclasses must implement
    the ``.create_basic_builder()`` method.

    Here is what this mixin brings to the model:

    1. Creates :class:`.ModelConfig` and :class:`.LayerBuilder` subclasses for the model
       type and attaches them to the model class as ``.Config`` and ``.Builder``
       attributes.
    2. Implements the :meth:`.create_builder` method, which wraps
       ``.create_basic_builder()`` to implement common functionality across models.
    3. Implements the ``<model_cls>.create()`` method, which is an
       alias of ``<model_cls>.Builder.build()``
    4. Decorates the model's initializer to collect usage telemetry.

    """

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls._should_update_signature():
            cls._update_create_builder_signature()
            cls._update_create_signature()
        cls.Config = _ModelConfig._create_subclass_from_model(cls)

    @classmethod
    def _should_update_signature(cls):
        """
        This predicate determine whether the signature for create_builder() should
        be updated
        """

        if "create_builder" in cls.__dict__:
            # model_yz = config_yz.create_model()
            _logger.debug(
                f"Skipping update to the signature of {cls.__name__}.create_builder() "
                "because this class overrides create_builder().  This may impact the "
                f"ModelConfig type for {cls}."
            )
            return False
        return True

    @staticmethod
    def _get_base_create_builder_signature():
        return _inspect.signature(ModelMixin.create_builder.__func__)

    @classmethod
    def create_builder(
        cls,
        *basic_builder_args,
        device: "OptionalDeviceOrString" = None,
        adapters: _Optional[_Dict[str, "ModelAdapter"]] = None,
        active_adapter: _Union[int, str] = 0,
        arch_optimizers: _Optional[_Dict[str, "ArchOptimizer"]] = None,
        pretrained: "LenientOptionalBool" = _OptionalBool.NOTSET,
        pretrained_path: _Optional[str] = None,
        freeze_params: "LenientOptionalBool" = _OptionalBool.NOTSET,
        dtype: "OptionalDtypeOrString" = None,
        metadata: _Optional["_ModelMetadata"] = None,
        **basic_builder_kwargs,
    ):
        """
        Creates and returns a configured instance of the model's
        :class:`~.layers.LayerBuilder` subclass.  This method adds hooks to the result
        of :meth:`cls.create_basic_builder` to add common functionality related to model
        instantiation.

        Args:
            device (:obj:`torch.device` or :obj:`str`, optional): The target device for
                model parameters and other state.  This is also the default device for
                adapter layers, which can be overridden in :obj:`.ModelAdapter`.
            adapters (:obj:`dict`, optional): A dictionary that maps :obj:`str` adapter
                IDs to instances of :obj:`.ModelAdapter`.  The returned builder will
                call these adapters to attach adapter layers (such as LoRA) to newly
                created models.
            active_adapter (:obj:`int` or `str`, optional): A variable indicates which
                adapter should be activated when the model is being instantiated.
                If the value is an integer, that stands for the index of the adapter.
                If the value is a string, it should be a valid adapter id.
            arch_optimizers (:obj:`dict`, optional): A dictionary that maps :obj:`str`
                architecture optimizer IDs to instances of
                :obj:`.ArchOptimizer`. The returned builder will apply these
                architecture optimizers (such as KVQuant) to newly created models.
            pretrained (:obj:`bool`): A flag for loading pre-trained weights.  The model
                only loads from pretrained paths when ``pretrained`` is ``True``.  This
                same flag also applies to the model adapters when initializing adapters.
            pretrained_path (:obj:`str`, optional): A pretrained checkpoint path for the
                backbone model's state dict.  This state should not include adapter
                state.
            freeze_params (:obj:`bool`): A flag for freezing parameters of the model
                (i.e., setting their ``requires_grad`` attributes to ``False``).  Note:
                this flag only applies to the backbone model, not adapter parameters,
                since adapters have their own ``freeze_params`` option.
            metadata (:obj:`ModelMetadata`): Additional information about the model,
                such as details regarding the model's origination.
            dtype (:obj:`torch.dtype`, optional): The parameter dtype.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """
        builder = super().create_builder(
            *basic_builder_args,
            device=device,
            adapters=adapters,
            active_adapter=active_adapter,
            arch_optimizers=arch_optimizers,
            pretrained=pretrained,
            pretrained_path=pretrained_path,
            freeze_params=freeze_params,
            dtype=dtype,
            **basic_builder_kwargs,
        )

        config = cls.Config(
            *basic_builder_args,
            device=device,
            adapters=adapters,
            active_adapter=active_adapter,
            arch_optimizers=arch_optimizers,
            pretrained=pretrained,
            pretrained_path=pretrained_path,
            freeze_params=freeze_params,
            dtype=dtype,
            metadata=metadata,
            **basic_builder_kwargs,
        )
        builder.post_hooks.register(_post_hooks.AttachConfigPostHook(config))

        if metadata is None:
            metadata = _ModelMetadata()
        builder.post_hooks.register(_post_hooks.AttachMetadataPostHook(metadata))

        return builder

    @property
    def config(self) -> _Union["_ModelConfig", None]:
        """The :obj:`ModelConfig` used to create the model."""
        return getattr(self, "_tamm_model_config", None)

    @config.setter
    def config(self, value: "_ModelConfig") -> None:
        self._tamm_model_config = value

    @property
    def metadata(self) -> _Union["_ModelMetadata", None]:
        """
        A :obj:`ModelMetadata` for the model (or ``None`` if this has not been set).
        """
        return getattr(self, "_tamm_model_metadata", None)

    @metadata.setter
    def metadata(self, value: "_ModelMetadata") -> None:
        self._tamm_model_metadata = value



"""
layers.feed_forward
^^^^^^^^^^^^^^^^^^^

This submodule implements general feed-forward layers.

.. autoclass:: tamm.layers.FeedForward
    :show-inheritance:
    :members: create_builder
"""

from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

from tamm import _adapters_v1
from tamm.layers import activation as _activation
from tamm.layers import dropout as _dropout
from tamm.layers import linear as _linear
from tamm.layers import norm as _norm
from tamm.layers import residual as _residual
from tamm.layers import sequential as _sequential
from tamm.layers.common import ConfigurableLayerMixin as _ConfigurableLayerMixin
from tamm.typing import ModuleOrBuilder as _ModuleOrBuilder
from tamm.typing import OptionalModuleOrBuilder as _OptionalModuleOrBuilder


class FeedForward(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    A simple :obj:`FeedForward` network with any number of hidden layers.  The output
    is the result of the following layer sequence:

    * Hidden transform 0
    * Activation 0
    * Activation dropout 0
    * Hidden transform 1
    * Activation 1
    * Activation dropout 1
    * ...
    * Output transform
    * Output dropout

    Args:
        * hidden_transforms: A :obj:`list` of hidden transform layers
            (:obj:`.LayerBuilder` or :obj:`torch.nn.Module`).
        * activations: A :obj:`list` of activation layer specifications.  Each element
            must be a valid argument to
            :func:`.layers.activation.create_activation_layer`.  The length of this
            arg must match the length of ``hidden_transforms``.
        * activation_dropouts: An optional :obj:`list` of activation dropout layers.
            Defaults to ``None``, which omits activation dropout.  If provided, the
            length must match the length of ``activations``.
        * output_transform: The output transform layer.
        * output_dropout: The output dropout layer.
    """

    def __init__(
        self,
        *,
        hidden_transforms: _List[_ModuleOrBuilder],
        activations: _List[_ModuleOrBuilder],
        activation_dropouts: _Optional[_List[_ModuleOrBuilder]] = None,
        output_transform: _ModuleOrBuilder,
        output_dropout: _OptionalModuleOrBuilder = None,
    ):
        self._validate_hidden_layers(
            hidden_transforms, activations, activation_dropouts
        )

        layers = {}
        for idx, hidden_transform in enumerate(hidden_transforms):
            layers[f"hidden_transform_{idx}"] = hidden_transform
            layers[f"activation_{idx}"] = _activation.create_activation_layer(
                activations[idx]
            )
            layers[f"activation_dropout_{idx}"] = (
                activation_dropouts[idx] if activation_dropouts is not None else None
            )
        layers["output_transform"] = output_transform
        layers["output_dropout"] = output_dropout

        super().__init__(layers, unpack_tuple_inputs=True)

    def _validate_hidden_layers(self, transforms, activations, dropouts):
        if len(transforms) != len(activations):
            raise ValueError("hidden_transforms and activations have different lengths")
        if dropouts is None:
            return
        if len(activations) != len(dropouts):
            raise ValueError(
                "acitvations and activation_dropouts have different lengths"
            )

    @classmethod
    def create_basic_builder(
        cls,
        *,
        input_dim: int,
        hidden_dim: _Optional[_Union[int, _List[int]]] = None,
        output_dim: _Optional[int] = None,
        num_layers: int = 2,
        activation: _activation.ActivationSpecType = "relu",
        hidden_transform_bias: bool = False,
        output_transform_bias: bool = False,
        activation_dropout_p: float = 0.0,
        output_dropout_p: float = 0.0,
    ):  # pylint: disable=arguments-differ,too-many-locals
        """
        Creates a builder for a simple :obj:`.FeedForward` layer with linear hidden
        and output transforms.  The hidden dimension and activation is consistent
        across layers.

        Args:
            input_dim (:obj:`int`): The feature dimension of the inputs.
            hidden_dim (:obj:`int` or :obj:`List[int]`):
                The dimensions of the hidden transform(s).
                If a list is provided, the length must equal to `num_layers-1`.
                If an integer is provided, this implies all the hidden layer(s) will
                have the same dimension.  If ``None``, all the hidden layer(s)
                will use ``input_dim`` as their dimension(s).
            output_dim (:obj:`int`): The dimension of the outputs.  If ``None``, this
                defaults to ``input_dim``.
            num_layers (:obj:`int`): The depth of the network, i.e., the number of
                ``hidden_transform`` layers plus one.  Defaults to ``2``.
            activation: The activation spec.  This should be a valid argument to
                :func:`.layers.activation.create_activation_layer`.
            hidden_transform_bias (:obj:`bool`): A flag for including a bias parameter
                in the hidden transforms.  Defaults to ``False``.
            output_transform_bias (:obj:`bool`): A flag for including a bias parameter
                in the output transform.  Defaults to ``False``.
            activation_dropout_p (:obj:`float`): The dropout parameter for activation
                tensors.  Defaults to ``0``.
            output_dropout_p (:obj:`float`): The dropout parameter for output
                tensors.  Defaults to ``0``.

        Returns:
            The configured :obj:`.LayerBuilder`.
        """
        if hidden_dim is None:
            hidden_dim = input_dim

        if isinstance(hidden_dim, int):
            hidden_dim = [hidden_dim] * (num_layers - 1)

        if not isinstance(hidden_dim, list):
            raise ValueError(
                "Unsupport type for optional argument ``hidden_dim``. "
                "Supported types are `int` and `List[int]`."
            )
        if len(hidden_dim) != num_layers - 1:
            raise ValueError(
                "Length of ``hidden_dim`` must be ``num_layers``-1. Got "
                f"len(hidden_dim)={len(hidden_dim)} and num_layers={num_layers}."
            )
        hidden_dims = hidden_dim

        if output_dim is None:
            output_dim = input_dim

        hidden_transforms = []
        activation_module = _activation.create_activation_layer(activation)
        for dim in hidden_dims:
            if activation_module.is_gated:
                hidden_transform = _linear.MultiOutputLinear.Builder(
                    input_dim,
                    2 * [dim],
                    bias=hidden_transform_bias,
                )
            else:
                hidden_transform = _linear.Linear.Builder(
                    input_dim, dim, bias=hidden_transform_bias
                )
            hidden_transforms.append(hidden_transform)
            input_dim = dim

        activations = [activation] * (num_layers - 1)

        activation_dropout = _dropout.Dropout.Builder(p=activation_dropout_p)
        activation_dropouts = [activation_dropout] * (num_layers - 1)

        output_transform = _linear.Linear.Builder(
            hidden_dims[-1] if len(hidden_dims) > 0 else input_dim,
            output_dim,
            bias=output_transform_bias,
        )

        output_dropout = _dropout.Dropout.Builder(p=output_dropout_p)

        return cls.Builder(
            hidden_transforms=hidden_transforms,
            activations=activations,
            activation_dropouts=activation_dropouts,
            output_transform=output_transform,
            output_dropout=output_dropout,
        )


class TransformerFeedForward(_sequential.Sequential, _ConfigurableLayerMixin):
    """
    Feed-forward layer with residual connection.  The output is the result of the
    following layer sequence:

    * Norm
    * Hidden transform
    * Activation
    * Dropout
    * Output transform
    * Dropout
    * Residual add

    Args:
        norm: The norm layer (:obj:`nn.Module`) or builder (callable that returns the
            norm layer when called without args).
        hidden_transform: The first transform layer or builder.
        activation: The activation layer or builder.  This can be any argument for
            ``tamm.layers.activation.create_activation_layer``.
        activation_dropout: The first dropout layer or builder.
        output_transform: The second transform layer or builder.
        output_dropout: The second dropout layer or builder.
        residual_connection: The residual connection layer of builder.


    Users can also create this layer using its :meth:`create` method, which has
    simpler args.  Alternatively, use :meth:`create_builder` to create the layer via
    a configurable builder.
    """

    def __init__(
        self,
        *,
        norm: _ModuleOrBuilder,
        hidden_transform: _ModuleOrBuilder,
        activation: _activation.ActivationSpecType,
        activation_dropout: _ModuleOrBuilder,
        output_transform: _ModuleOrBuilder,
        output_dropout: _ModuleOrBuilder,
        residual_connection: _ModuleOrBuilder,
        adapter_target_prefix_str: str,
    ):
        self.adapter_target_prefix_str = adapter_target_prefix_str
        activation = _activation.create_activation_layer(activation)
        named_layers = {
            "norm": norm,
            "hidden_transform": hidden_transform,
            "activation": activation,
            "activation_dropout": activation_dropout,
            "output_transform": output_transform,
            "output_dropout": output_dropout,
        }
        super().__init__(
            named_layers,
            unpack_tuple_inputs=True,
            residual_connection=residual_connection,
        )
        self._mark_adaptable_layers()

        # Segmented linear option accepts group_sizes side inputs
        self.register_side_input("hidden_transform", "group_sizes")
        self.register_side_input("output_transform", "group_sizes")

    @classmethod
    def create_basic_builder(
        cls,
        *,
        input_dim: int,
        hidden_dim: int,
        norm: _Union[str, _OptionalModuleOrBuilder] = "layer_norm",
        activation: _Optional[_activation.ActivationSpecType] = None,
        activation_dropout_p: float = 0.0,
        hidden_transform_bias: bool = False,
        output_transform_bias: bool = False,
        output_dropout_p: float = 0.0,
        apply_residual_add: bool = True,
        vec_dim: _Optional[_List[int]] = None,
        adapter_target_prefix_str: str = "",
        use_segmented_linear: bool = False,
    ):  # pylint: disable=arguments-differ,too-many-locals
        """
        Creates and returns a default builder for creating :obj:`TransformerFeedForward`
        objects.  The builder uses a :class:`LayerNorm` norm layer, linear transforms,
        and ReLU activation.

        Use :meth:`create` to directly create a feed forward layer.

        Args:
            input_dim (:obj:`int`): The input dimension.  This should equal the last
                value of the input shape.
            hidden_dim (:obj:`int`): The hidden dimension.
            norm: The norm layer (:obj:`nn.Module`) or builder (a callable that returns
                the norm layer when called without args).  If ``None``, this defaults to
                :class:`LayerNorm`.
            activation: The activation layer or builder.  This can be any argument for
                ``tamm.layers.activation.create_activation_layer``.  If ``None``, this
                defaults to :class:`ReLU`.
            activation_dropout_p (:obj:`float`): The dropout probability for
                activations.
            hidden_transform_bias (:obj:`bool`): Whether the hidden transform layers
                uses a bias or not.
            output_transform_bias (:obj:`bool`): Whether the output transform layer
                uses a bias or not.
            output_dropout_p (:obj:`float`): The dropout probability for outputs (prior
                to the residual add).
            apply_residual_add (:obj:`bool`): Whether the feed forward layer uses a
                residual connection or not.

        Returns:
            A configured :obj:`TransformerFeedForward.Builder`.
        """

        # High level args for vectorized linear / linear kwargs
        multioutput_linear_kwargs = {}
        linear_cls = _linear.Linear
        linear_kwargs = {}
        if vec_dim is not None:
            multioutput_linear_kwargs["vec_dim"] = vec_dim
            linear_kwargs["vec_dim"] = vec_dim
            linear_cls = _linear.VectorizedLinear

        # For segmented linear, linear class is segmented gemm, with vectorized dimension inputs
        if use_segmented_linear:
            linear_cls = _linear.SegmentedLinear
            multioutput_linear_kwargs["override_linear_cls"] = _linear.SegmentedLinear

        norm = _norm.create_norm_builder((input_dim,), spec=norm)
        if activation is None:
            activation = _activation.ReLU.Builder()

        activation_module = _activation.create_activation_layer(activation)
        if activation_module.is_gated:
            hidden_transform = _linear.MultiOutputLinear.Builder(
                input_dim,
                2 * [hidden_dim],
                bias=hidden_transform_bias,
                **multioutput_linear_kwargs,
            )
        else:
            hidden_transform = linear_cls.Builder(
                input_dim, hidden_dim, bias=hidden_transform_bias, **linear_kwargs
            )
        activation_dropout = _dropout.Dropout.Builder(p=activation_dropout_p)

        output_transform = linear_cls.Builder(
            hidden_dim,
            input_dim,
            bias=output_transform_bias,
            **linear_kwargs,
        )
        output_dropout = _dropout.Dropout.Builder(p=output_dropout_p)

        residual_connection = (
            _residual.ResidualAdd.Builder() if apply_residual_add else None
        )

        return cls.Builder(
            norm=norm,
            hidden_transform=hidden_transform,
            activation=activation,
            activation_dropout=activation_dropout,
            output_transform=output_transform,
            output_dropout=output_dropout,
            residual_connection=residual_connection,
            adapter_target_prefix_str=adapter_target_prefix_str,
        )

    def _mark_adaptable_layers(self):
        output_transform_layer_type = (
            "BatchedFeedForwardOutputTransform"
            if isinstance(self.output_transform, _linear.VectorizedLinear)
            else "FeedForwardOutputTransform"
        )
        _adapters_v1.annotate_layer(
            self.output_transform,
            [(output_transform_layer_type,)],
        )

        layers = (
            [self.hidden_transform.linear_0, self.hidden_transform.linear_1]
            if self.activation.is_gated
            else [self.hidden_transform]
        )

        for layer in layers:
            layer_type = (
                "BatchedFeedForwardHiddenTransform"
                if isinstance(layer, _linear.VectorizedLinear)
                else "FeedForwardHiddenTransform"
            )
            _adapters_v1.annotate_layer(
                layer,
                [(layer_type,)],
            )



"""
.. autoclass:: tamm.layers.common.LayerBuilder
    :members:
    :show-inheritance:

"""
import logging as _logging

from tamm import _helpers
from tamm.utils import partial as _partial
from tamm.utils.json import JSONSerializableMixin as _JSONSerializableMixin

_logger = _logging.getLogger(__name__)


class LayerBuilder(
    _partial.DataclassedPartial,
    _JSONSerializableMixin,
    json_namespace="layers",
):
    """
    The base class for layer builders.  A :class:`.LayerBuilder` is a
    :class:`.DataclassedPartial` with a |tamm| layer class as the target callable.
    """

    def _to_json_dict_impl(self):
        result = _helpers.dataclass_to_dict(
            self.configured_args,
            omit_defaults=True,  # only include non-defaults for forward compatibility
        )
        return result

    def build(self, *override_args, **override_kwargs):
        """
        Creates and returns a configured instance of the layer.

        Args:
            override_args: Optional positional arguments to override args specified in
                the builder.  These args replace the first ``len(override_args)``
                positional args (and *all* varargs if ``override_args`` contains
                varargs) in the layer's constructor.
            override_kwargs: Optional keyword override arguments.  These arguments
                replace any additional named arguments not overriden by
                ``override_args``.

        Returns:
            The newly created layer.
        """
        return self(*override_args, **override_kwargs)


class BuildableMixin:
    """
    A mixin to attach ``.Builder`` to a class.
    This mixin is meant to supersede the ``@buildable`` decorator.
    """

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        _buildable_no_deprecate(cls)


def _buildable_no_deprecate(layer_cls):
    """
    Creates a :class:`.LayerBuilder` for a layer class and attaches the newly created
    type to ``layer_cls`` as the ``Builder`` attribute.
    """
    layer_name = layer_cls.__name__
    builder_name = f"{layer_name}Builder"
    builder_cls = LayerBuilder.create_subclass(
        target_callable=layer_cls, name=builder_name
    )
    builder_cls.__doc__ = (
        f"A :class:`.LayerBuilder` subclass for configuring :class:`.{layer_name}` "
        f"layers.  Use the alias :attr:`.{layer_name}.Builder` to access this class. "
        f"Please check :class:`.{layer_name}` for more details about the signature."
    )
    layer_cls.Builder = builder_cls
    return layer_cls



"""
transformer.kv_cache
^^^^^^^^^^^^^^^^^^^^

This module implements key-value cache components to speed up streaming inference with
causal transformer stacks.  Language model text generation is an important example of
this.


Base classes
~~~~~~~~~~~~

These classes define the interface for KV caching.

.. autoclass:: tamm.layers.transformer.BaseKVCache
    :members:

.. autoclass:: tamm.layers.transformer.BaseKVCacheView
    :members:

.. autoclass:: tamm.layers.transformer.KVCacheLayerView
    :members:


Vanilla cache implementation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These classes provide a simple cache implementation that covers basic inference
scenarios.

.. autoclass:: tamm.layers.transformer.VanillaKVCache
    :members:
    :show-inheritance:

.. autoclass:: tamm.layers.transformer.VanillaKVCacheView
    :members:
    :show-inheritance:
"""

from tamm.layers.transformer.kv_cache.common import (
    BaseKVCache,
    BaseKVCacheView,
    KVCacheLayerView,
)
from tamm.layers.transformer.kv_cache.speculative import (
    SpeculativeKVCache,
    SpeculativeKVCacheView,
)
from tamm.layers.transformer.kv_cache.v0 import V0KVCache, V0KVCacheView
from tamm.layers.transformer.kv_cache.vanilla import VanillaKVCache, VanillaKVCacheView



from typing import List as _List
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch

from tamm.layers.transformer.kv_cache import common as _common


class V0KVCache(_common.BaseKVCache):
    """
    A :class:`KVCache` that is backward compatible with early versions of |tamm|.
    """

    def __init__(
        self,
        layers: int,
        batch: int,
        init_seq: int,
        dim: int,
        offsets: _Union[_List[int], _torch.Tensor],
        device=None,
        dtype=None,
    ):
        self.cache = _torch.empty(
            layers, 2, batch, init_seq, dim, device=device, dtype=dtype
        )
        if not isinstance(offsets, _torch.Tensor):
            offsets = _torch.tensor(offsets, device=device, dtype=_torch.int32)

        self.offsets = offsets

    @property
    def length(self) -> int:
        return self.cache.size(3)

    def up_to_index(self, index: int) -> "V0KVCacheView":
        return V0KVCacheView(cache=self.cache[:, :, :, :index], offsets=self.offsets)

    def resize(self, new_length: int = None, *, new_seq: int = None):
        if new_seq is None:
            new_seq = new_length

        layers, two, batch, old_seq, dim = self.cache.shape
        new_cache = _torch.empty(
            layers,
            two,
            batch,
            new_seq,
            dim,
            device=self.cache.device,
            dtype=self.cache.dtype,
        )
        new_cache[:, :, :, :old_seq, :] = self.cache
        self.cache = new_cache

    def to(self, *args, **kwargs):  # pylint: disable=invalid-name
        self.cache = self.cache.to(*args, **kwargs)
        self.offsets = self.offsets.to(device=self.cache.device)
        return self


class V0KVCacheView(_common.BaseKVCacheView):
    """
    A :class:`BaseKVCacheView` that is backward compatible with early versions of
    |tamm|.  This cache requires the user to specify the amount of left padding for
    each sequence using the :attr:`offsets` tensor.

    Args:
        cache (:obj: `torch.Tensor): A tensor with shape layers x 2 x batch x seq x dim
        offsets (:obj: `torch.Tensor): An integer tensor with shape batch
    """

    def __init__(self, cache, offsets):
        self.cache = cache
        self.offsets = offsets

    @property
    def length(self) -> int:
        return self.cache.size(3)

    @property
    def num_layers(self) -> int:
        return self.cache.size(0)

    @property
    def dtype(self) -> _torch.dtype:
        return self.cache.dtype

    @property
    def device(self) -> _torch.device:
        return self.cache.device

    def at_layer(
        self, layer_idx: _Optional[int] = None, *, layer: _Optional[int] = None
    ) -> _common.KVCacheLayerView:
        if layer is None:
            layer = layer_idx
        return super().at_layer(layer)

    def write_to_tail(
        self, key: _torch.Tensor, value: _torch.Tensor, *, layer_idx: int
    ) -> None:
        self.cache[layer_idx, 0, :, -key.size(1) :, :] = key
        self.cache[layer_idx, 1, :, -value.size(1) :, :] = value

    def write_to_tail_vectorized(
        self, keys: _torch.Tensor, values: _torch.Tensor, *, layer_indices: _List[int]
    ) -> None:
        # Vectorized implementation of write_to_tail
        keys = keys.transpose(0, 1)
        values = values.transpose(0, 1)

        if len(layer_indices) != keys.shape[0] or len(layer_indices) != values.shape[0]:
            raise ValueError("Mismatched V0KVCacheLayersView vectorize dimension")

        # Vectorized write
        self.cache[layer_indices, 0, :, -keys.size(2) :, :] = keys.reshape(
            (len(layer_indices), keys.size(1), keys.size(2), keys.size(3))
        )
        self.cache[layer_indices, 1, :, -values.size(2) :, :] = values.reshape(
            (len(layer_indices), keys.size(1), values.size(2), values.size(3))
        )

    def write_segment_ids_to_tail(self, segment_ids: _torch.Tensor) -> None:
        pass

    def write_positions_to_tail(self, positions: _torch.Tensor) -> None:
        pass

    def read(self, layer_idx: int) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        return self.cache[layer_idx, 0], self.cache[layer_idx, 1]

    def read_vectorized(
        self, layer_indices: _List[int]
    ) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        # Vectorized implementation of read_multi
        qs = self.cache[layer_indices, 0]
        vs = self.cache[layer_indices, 1]

        # Swap axes -> ( batch, track, ...)
        qs = qs.transpose(0, 1)
        vs = vs.transpose(0, 1)

        return qs, vs

    def read_segment_ids(self) -> _torch.Tensor:
        kv_positions = _torch.arange(self.length, device=self.device)
        return self.offsets[:, None] <= kv_positions[None, :]

    def read_positions(self) -> _torch.Tensor:
        kv_positions = _torch.arange(self.length, device=self.device)

        batch_size = self.cache.size(2)
        kv_positions = kv_positions[None, :].tile(batch_size, 1)

        return _torch.where(
            kv_positions < self.offsets[..., None],
            kv_positions,
            kv_positions - self.offsets[..., None],
            # For each sequence, the first offsets[idx] tokens are padding, so we
            # subtract this value to ensure that the non-padding positions start form 0.
            # We treat the paddings like another segment with positions 0, 1, 2, ...
        )



from torch import nn as _nn

from tamm import context_vars as _context_vars
from tamm.layers.common._marker import get_marker, update_marker
from tamm.utils import OptionalBool as _OptionalBool


def _should_update_requires_grad(model: "_nn.Module") -> bool:
    """
    Check ._tamm_marker of a Module to determine whether to actually
    freeze the parameter. If .tamm_metadata is set to an explicit
    ``freeze_params_hook_applied`` value (either True or False),
    skip this post hook.
    """
    try:
        if get_marker(model).freeze_params_hook_applied != _OptionalBool.NOTSET:
            return False
    except AttributeError:
        return True
    return True


def require_explicit_freeze_params(func):
    """ "
    Decorate a function with this to raise an error when freeze_params is NOTSET.
    """

    def wrapper(module: "_nn.Module", freeze_params=_OptionalBool.NOTSET):
        if freeze_params == _OptionalBool.NOTSET:
            raise ValueError("freeze_params must be TRUE or FALSE")
        func(module, freeze_params)

    return wrapper


@require_explicit_freeze_params
def _recursively_freeze_params(
    module: "_nn.Module", freeze_params=_OptionalBool.NOTSET
):
    if not _should_update_requires_grad(module):
        return
    _freeze_direct_params(module, freeze_params=freeze_params)
    for sub_module in module.children():
        _recursively_freeze_params(sub_module, freeze_params=freeze_params)


@require_explicit_freeze_params
def _freeze_direct_params(module: "_nn.Module", freeze_params=_OptionalBool.NOTSET):
    update_marker(module, freeze_params_hook_applied=freeze_params)
    for param in module.parameters(recurse=False):
        param.requires_grad = not bool(freeze_params)


class FreezeParamsPostHook:
    def __call__(self, model: "_nn.Module") -> None:
        resolved = _context_vars.resolve_freeze_params()
        if resolved == _OptionalBool.NOTSET:
            return
        _recursively_freeze_params(model, freeze_params=resolved)

    def __repr__(self):
        return f"{self.__class__.__name__}()"



import math as _math
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch


def shape_normalized_normal_(
    tensor: _torch.Tensor,
    *,
    dim: _Optional[_Union[int, _Tuple[int, ...]]] = None,
):
    """
    Fills a tensor with values drawn i.i.d. from a zero-centered normal
    distribution.  The variance is defined such that the norm of these values
    along ``dim`` has an expected value of ``1``.

    Args:
        tensor (:obj:`torch.Tensor`): The tensor to initialize.
        dim (:obj:`int` or :obj:`tuple` of :obj:`int`, optional):  The
            dimension(s) of ``tensor`` to consider when choosing the variance of
            the initialization distribution.  Defaults to all dimensions.
    """

    if dim is None:
        count = tensor.numel()
    elif isinstance(dim, int):
        count = tensor.size(dim)
    else:
        sizes = [tensor.size(d) for d in dim]
        count = _torch.Size(sizes).numel()

    std = 1 / _math.sqrt(count)
    _torch.nn.init.normal_(tensor, std=std)



import abc as _abc
import dataclasses as _dataclasses
import textwrap as _textwrap
from typing import Optional as _Optional
from typing import Set as _Set

import torch as _torch
import torch.nn as _nn

from tamm import _helpers
from tamm.context_vars import resolve_pretrained_flag as _resolve_pretrained_flag
from tamm.layers.common import PretrainedLoader as _PretrainedLoader
from tamm.typing import LenientOptionalBool
from tamm.typing import OptionalDeviceOrString as _OptionalDeviceOrString
from tamm.utils import OptionalBool as _OptionalBool
from tamm.utils.json import JSONSerializableABCMixin

# pylint: disable=no-member  # pylint does not pick up some dataclass fields
# pylint: disable=duplicate-code  # probably should refactor with ModelAdapter


class ArchOptimizer(JSONSerializableABCMixin, json_namespace="ao"):
    """
    A base class for objects that apply optimizations to |tamm| models.
    """

    # pylint: disable=signature-differs
    def __init_subclass__(cls):
        cls._inject_common_fields_and_make_dataclass()

    @classmethod
    def _inject_common_fields_and_make_dataclass(cls):
        _helpers.set_annotated_class_attr(cls, "pretrained_path", None, _Optional[str])
        _helpers.set_annotated_class_attr(cls, "device", None, _OptionalDeviceOrString)
        _dataclasses.dataclass(cls, repr=False)

    def optimize(
        self,
        model: _nn.Module,
        *,
        pretrained: "LenientOptionalBool" = _OptionalBool.NOTSET,
    ) -> None:
        """
        Applies optimizations to a model.

        Args:
            model (:obj:`torch.nn.Module`): The model to optimize.
            pretrained (:obj:`LenientOptionalBool`): A flag for loading pretrained
                state. If ``OptionalBool.TRUE``, then the method loads the state dict
                from ``pretrained_path`` into newly created layers. Python ``True``,
                and ``False`` are also accepted.
        """
        resolved_pretrained_flag = _resolve_pretrained_flag(pretrained)
        should_load_pretrained = (
            resolved_pretrained_flag == _OptionalBool.TRUE
            and self.pretrained_path is not None
        )
        target_device = self.device
        if should_load_pretrained:
            target_device = "meta"

        prior_state = set(_helpers.get_all_params_and_buffers(model))

        with _helpers.get_device_context(target_device):
            self._optimize_impl(model, pretrained=resolved_pretrained_flag)

        if should_load_pretrained:
            self._load_pretrained(model=model, prior_state=prior_state)

    def _load_pretrained(
        self, *, model: _nn.Module, prior_state: _Set[_torch.Tensor]
    ) -> None:
        if self.pretrained_path is None:
            return

        new_keys = {
            name
            for name, tensor in model.state_dict(keep_vars=True).items()
            if tensor not in prior_state
        }
        loader = _PretrainedLoader(
            path=self.pretrained_path,
            keys=new_keys,
            target_device=self.device,
        )
        loader.load(model)

    @_abc.abstractmethod
    def _optimize_impl(
        self,
        model: _nn.Module,
        *,
        pretrained: "_OptionalBool" = _OptionalBool.NOTSET,
    ) -> None:
        """Applies optimizations to a model."""

    def _to_json_dict_impl(self):
        return _helpers.dataclass_to_dict(
            self, omit_defaults=True  # only non-defaults for forward compatibility
        )

    def __repr__(self):
        args_dict = _helpers.dataclass_to_dict(self)
        args_reprs = [f"{key}={repr(value)}," for key, value in args_dict.items()]
        joined_args = "\n".join(args_reprs)
        intended_args = _textwrap.indent(joined_args, prefix=" " * 4)
        return f"{self.__class__.__name__}(\n{intended_args}\n)"



import json as _json
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Optional as _Optional

from tamm import tokenizers as _tokenizers
from tamm.preprocessors.instruct_lm import jinja as _jinja
from tamm.preprocessors.instruct_lm.afm import utils as _utils

DEFAULT_SYSTEM_MESSAGE = "A conversation between a user and a helpful assistant."


V6_JINJA_CHAT_TEMPLATE = """
{{- '<turn_start> ' + messages[0]['role'] + '<n>' + messages[0]['content'] -}}
{% if tools %}
    {{- ('<n>system tools: ' + (tools | map('tojson') | join('<n>'))) -}}
{% endif %}
{{- '<turn_end>' -}}
{% for message in messages[1:] %}
    {{- '<turn_start> ' + message['role'] + '<n>' + message['content'] + '<turn_end>' -}}
{% endfor %}
{% if add_generation_prompt is defined and add_generation_prompt %}
    {% if messages[-1]['role'] != target_role %}
        {{- '<turn_start> ' + target_role + '<n>' -}}
    {% endif %}
{% endif %}
""".strip()


class AFMChatTemplateV6Preprocessor(_jinja.JinjaChatTemplatePreprocessor):
    def __init__(
        self,
        tokenizer_spec: _tokenizers.TokenizerSpecType,
        generate_prefix: bool = False,
        max_sequence_length: _Optional[int] = None,
        prepend_system_message: bool = True,
        system_message: str = DEFAULT_SYSTEM_MESSAGE,
        target_role: str = "assistant",
        ignored_id: int = -100,
    ):
        super().__init__(
            tokenizer_spec=tokenizer_spec,
            chat_template=V6_JINJA_CHAT_TEMPLATE,
            generate_prefix=generate_prefix,
            max_sequence_length=max_sequence_length,
            prepend_system_message=prepend_system_message,
            system_message=system_message,
            strip_first_token=True,
            prepend_eos=True,
            target_role=target_role,
            ignored_id=ignored_id,
        )

    def maybe_update_message(self, message: _Dict[str, _Any]) -> _Dict[str, _Any]:
        """
        Updates the message to add tool calling and response format capabilities.
        """

        tool_calls = message.pop("tool_calls", None)
        if tool_calls is not None:
            tool_calls = _utils.tool_calls_to_json_str(tool_calls)
            message[
                "content"
            ] += f"<executable_start> ```function\n{tool_calls}```<executable_end>"

        if message["role"] in ("tool", "function"):
            message["role"] = "tool"
            content = message.pop("content")
            if isinstance(content, dict):
                # tool response case
                content = _json.dumps(content, ensure_ascii=False)
            elif isinstance(content, list):
                # newly added tools case
                tools = _utils.tools_to_json_str(content)
                content = f"new tools: {tools}"
            message["content"] = content

        response_format = message.pop("response_format", None)
        if response_format is not None:
            if response_format["type"] != "json_schema":
                raise ValueError(f"Unrecognized response format {response_format}")
            message["content"] += self._response_format_to_str(response_format)

        return message

    @staticmethod
    def _response_format_to_str(response_format: _Dict[str, _Any]) -> str:
        json_schema = response_format["json_schema"]

        pieces = ["\n\nresponse format in json.\n"]

        for key in ["name", "description"]:
            value = json_schema.get(key)
            pieces.append(f"{key}: {value}\n")

        schema = json_schema["schema"]
        schema_value = _json.dumps(schema, ensure_ascii=False)
        pieces.append(f"schema: {schema_value}")

        return "".join(pieces)



import abc
import copy
import logging
from itertools import chain
from typing import Any, List, Sequence, Union

from tamm.tokenizers.common import Tokenizer, TokenizerConfig

logger = logging.getLogger(__name__)


def validate_tokenizer_identifier(identifier: Any):
    if identifier is None or identifier == "":
        raise ValueError(
            f"TokenizerRegistry().create_config() requires an identifier, got {identifier}"
        )
    if isinstance(identifier, (int, float, bool)):
        raise ValueError(f"Tokenizer identifier cannot be {type(identifier).__name__}")


class TokenizerRegistry(abc.ABC):
    def __init__(self, *args, **kwargs):
        del args, kwargs

    def create_config(
        self, identifier: Union[str, "TokenizerConfig"], *args, **kwargs
    ) -> "TokenizerConfig":
        """
        Creates a Tokenizer referred by ``identifier``

        Args:
            identifier: :class:`TokenizerConfig` or the name of the Tokenizer
            **kwargs: Extra keyword arguments

        Returns:
            Tokenizer config instance
        """
        validate_tokenizer_identifier(identifier)
        if isinstance(identifier, TokenizerConfig):
            tokenizer = identifier
        else:
            if not isinstance(identifier, str):
                raise ValueError(
                    f"Expected a string tokenizer identifier, got '{type(identifier).__name__}'"
                )
            tokenizer = self._create_config(identifier, *args, **kwargs)
        # always use a shallow copy so that instances do not share configurable attributes
        tokenizer = copy.copy(tokenizer)
        for k, v in kwargs.items():
            setattr(tokenizer, k, v)
        return tokenizer

    def create(
        self, identifier: Union[str, "Tokenizer"], *args, **kwargs
    ) -> "Tokenizer":
        """
        Creates a Tokenizer referred by ``identifier``

        Args:
            identifier: :class:`Tokenizer` or the name of the Tokenizer
            **kwargs: Extra keyword arguments

        Returns: Tokenizer instance
        """
        validate_tokenizer_identifier(identifier)

        if isinstance(identifier, Tokenizer):
            return identifier

        tokenizer_config = self.create_config(identifier, *args, **kwargs)
        return tokenizer_config.create_tokenizer(**kwargs)

    def clear_cache(self):
        """
        Clears any local cache if setup by the concrete registry
        """

    @abc.abstractmethod
    def _create_config(
        self, identifier: Union[str, "Tokenizer"], *args, **kwargs
    ) -> "TokenizerConfig":
        """
        Abstract method for creating a Tokenizer.
        Shall be implemented by concrete class.
        """

    @abc.abstractmethod
    def list_(self, *args, **kwargs) -> List[str]:
        """
        Returns a list of Tokenizers available

        Args:
            *args: Extra positional arguments
            **kwargs: Extra keyword arguments

        Returns:
             list of Tokenizer names
        """

    @abc.abstractmethod
    def describe(self, identifier: str, *args, **kwargs) -> str:
        """
        Describe a tokenizer

        Args:
            identifier: name of the tokenizer
            *args: Extra positional arguments
            **kwargs: Extra keyword arguments

        Returns:
             Description of the tokenizer
        """

    def __add__(self, other: "TokenizerRegistry") -> "TokenizerRegistry":
        """
        Combines two TokenizerRegistryBase instances into a composite registry.

        The Tokenizers defined in `self` take precedence over those in `other`
        for the same `model_id`.

        Args:
            other (:obj: `TokenizerRegistryBaseBase`): Another Tokenizer registry to combine.

        Returns:
            :obj: `TokenizerRegistryBaseBase`: A composite registry combining `self` and `other`.
        """
        return ComposedTokenizerRegistry(self, other)

    def __iadd__(self, other: "TokenizerRegistry") -> "TokenizerRegistry":
        """
        In place version if __add__

        Args:
            other (:obj: `TokenizerRegistryBase`): Another Tokenizer registry to combine.

        Returns:
            :obj: `TokenizerRegistryBase`: A composite registry combining `self` and `other`.
        """
        return ComposedTokenizerRegistry(self, other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(...)"


class ComposedTokenizerRegistry(TokenizerRegistry):
    """
    Represents a composite registry combining multiple TokenizerRegistryBase instances.
    """

    def __init__(self, *tokenizer_registries: "TokenizerRegistry"):
        super().__init__()
        self._registries: Sequence["TokenizerRegistry"] = tokenizer_registries

    def clear_cache(self) -> None:
        for registry in self._registries:
            registry.clear_cache()

    def _create_config(
        self, identifier: Union[str, "TokenizerConfig"], *args, **kwargs
    ) -> "TokenizerConfig":
        for registry in self._registries:
            try:
                return registry.create_config(identifier, *args, **kwargs)
            except Exception as e:
                logger.debug("Cannot create %s with %s: %s", identifier, registry, e)
        raise KeyError(
            f"{identifier} is not found in all available tokenizer registries: {self._registries}"
        )

    def list_(self, *args, **kwargs) -> List[str]:
        iterators = [registry.list_(*args, **kwargs) for registry in self._registries]
        return list(chain(*iterators))

    def describe(self, identifier: str, *args, **kwargs) -> str:
        for registry in self._registries:
            try:
                return registry.describe(identifier, *args, **kwargs)
            except Exception as e:
                logger.debug("%ss cannot describe %s: %s", registry, identifier, e)
        raise KeyError(
            f"{identifier} cannot be described by any tokenizer registry: {self._registries}"
        )

    def __repr__(self):
        composed = ", ".join([str(registry) for registry in self._registries])
        return f"{self.__class__.__name__}({composed})"



import logging as _logging
from typing import Set as _Set

import torch.nn as _nn

from tamm.context_vars import get_model_build_device as _get_model_build_device
from tamm.context_vars import resolve_pretrained_flag as _resolve_pretrained_flag
from tamm.layers.common._marker import get_marker, update_marker
from tamm.layers.common.pretrained_loader import PretrainedLoader
from tamm.utils import OptionalBool

_logger = _logging.getLogger(__name__)


class ModelInitializerPostHook:
    """
    Applies pretrained context if explicitly provided in
    ``create_builder(pretrained=[True|False])`` calls.
    """

    def __init__(self, path: str):
        self.path = path

    def __call__(self, model: _nn.Module) -> None:
        if _resolve_pretrained_flag() == OptionalBool.TRUE and self.path is None:
            return

        if _resolve_pretrained_flag() == OptionalBool.FALSE:
            # Users specify to random initialize the current module.
            update_marker(model, weights_initialized=True)
            return

        target_keys = self._get_target_keys(model)
        PretrainedLoader(
            path=self.path,
            target_device=_get_model_build_device(disable_meta_init_trick=True),
            keys=target_keys,
        ).load(model)
        update_marker(model, weights_initialized=True)

    @classmethod
    def _get_target_keys(
        cls, module, target_keys: _Set[str] = None, prefix: str = ""
    ) -> _Set[str]:
        if target_keys is None:
            target_keys = set()

        # Store buffers
        target_keys = target_keys.union(
            # pylint: disable=protected-access
            {
                prefix + ("." if prefix else "") + name
                for name, _ in module.named_buffers(recurse=False)
                if name not in module._non_persistent_buffers_set
            }
        )

        # Store parameters
        target_keys = target_keys.union(
            {
                prefix + ("." if prefix else "") + name
                for name, _ in module.named_parameters(recurse=False)
            }
        )

        # Loop through all submodules
        for name, sub_module in module.named_children():
            if get_marker(sub_module).weights_initialized:
                continue
            submodule_prefix = prefix + ("." if prefix else "") + name
            target_keys = target_keys.union(
                cls._get_target_keys(sub_module, target_keys, submodule_prefix)
            )
        return target_keys



"""
layers.rope
^^^^^^^^^^^

This submodule provides utilities for
`rotary positional embeddings <https://arxiv.org/abs/2104.09864>`__.

.. autofunction:: tamm.layers.rope.compute_rope_coefficients

.. autofunction:: tamm.layers.rope.rotate
"""

import math as _math
from typing import Optional as _Optional
from typing import Tuple as _Tuple

import torch as _torch


def compute_rope_coefficients(
    *,
    positions: _torch.Tensor,
    dim: int,
    theta: float = 10000.0,
    dtype: _Optional[_torch.dtype] = None,
) -> _torch.Tensor:
    """
    Computes sine and cosine values for rotating a tensor with :func:`~.rope.rotate`.

    Args:
        positions (:obj:`torch.Tensor`): A ``batch_size x seq_length`` tensor of
            token positions.
        dim (:obj:`int`): The last dim of the tensor to rotate.  This must be divisible
            by 2.
        theta (:obj:`float`): The base value for computing sinusoid frequencies.
        dtype (:obj:`torch.dtype`): The dtype for the result.

    Returns:
        A tensor of sine and cosine values with shape
        ``[batch_size x seq_length, 1, (dim / 2),  2,  2]``.  The last two dimensions
        hold 2D rotation matrices.
    """
    if dim % 2:
        raise ValueError(f"dim has value {dim}, but this must be an even integer")

    # flatten batch_size and seq_length dims to support on-device export
    # (as of april 2025, coremltools limits tensors to 5 dimensions)
    positions = positions.flatten()

    # This is to avoid a bug in torch compile while ensuring numerical accuracy
    # Moving a small list of floats shouldn't create too much burden to GPUs
    exponents = [x * (2 / dim - 1) / (dim // 2 - 1) for x in range(dim // 2)]
    freqs_list = [_math.pow(theta, x) for x in exponents]
    freqs = _torch.tensor(freqs_list, dtype=_torch.float32, device=positions.device)
    angles = positions[..., None, None] * freqs
    sin, cos = _torch.sin(angles), _torch.cos(angles)
    if dtype is not None:
        sin, cos = sin.type(dtype), cos.type(dtype)
    result = _torch.stack([cos, -sin, sin, cos], dim=-1)
    return result.reshape(
        *result.shape[:-1], 2, 2
    )  # the last 2 dimensions are the 2d rotation matrices


def rotate(
    x: _torch.Tensor, *, rope_coefficients: _torch.Tensor
) -> _Tuple[_torch.Tensor, _torch.Tensor]:
    """
    Returns a rotation of ``x``.

    Args:
        x (:obj:`torch.Tensor`): A ``batch_size x seq_length x num_heads x dim`` tensor.
        rope_coefficients (:obj:`torch.Tensor`): A
            ``[batch_size x seq_length, 1, (dim / 2),  2,  2]`` tensor returned by
            :func:`compute_rope_coefficients`.

    Returns:
        The rotated ``x``.  This is a 2D rotation of each pair of values in the last
        dimension.
    """
    input_shape = x.shape
    input_dtype = x.dtype

    if len(x.shape) != 4:
        x = x.reshape(*input_shape[:-1], -1, 1, 2)
        # split batch_size and seq_length from the first dim
        rope_coefficients = rope_coefficients.reshape(
            input_shape[0], -1, *rope_coefficients.shape[1:]
        )
        # Make it work for a vectorized dimension (dimension after batch) (TODO: is there a better way)
        rope_coefficients = rope_coefficients.unsqueeze(1)
    else:
        # flatten batch_size and seq_length to match rope_coefficients
        x = x.reshape(-1, *input_shape[2:-1], input_shape[-1] // 2, 1, 2)

    # pylint: disable-next=not-callable
    x = _torch.linalg.vecdot(x, rope_coefficients, dim=-1)

    x = x.reshape(input_shape)
    return x.type(input_dtype)



"""
Prefix Tuning Adapter Implementation.
Ref paper: https://arxiv.org/pdf/2101.00190.pdf.
"""

from typing import Any as _Any
from typing import Optional as _Optional
from typing import Tuple
from typing import Union as _Union

import torch as _torch
import torch.nn as _nn

from tamm._adapters_v1.layer_adapters.common import LayerAdapter as _LayerAdapter
from tamm._adapters_v1.layer_adapters.common import (
    attach_config_class as _attach_config_class,
)


class MLP(_nn.Sequential):
    """
    MLP class for adjusting prefix dimension.
    Ideally the input prefix dimension should be smaller than embedding dimension.
    Then we use this MLP to project prefix embedding to a higher dimension
    which matches the embedding dimension.
    """

    def __init__(
        self, input_dim, output_dim, hidden_dim=512, dtype=_torch.float32, device=None
    ):
        super().__init__(
            _nn.Linear(input_dim, hidden_dim, bias=True, dtype=dtype, device=device),
            _nn.Tanh(),
            _nn.Linear(hidden_dim, output_dim, bias=True, dtype=dtype, device=device),
        )
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim

    def __repr__(self):
        return (
            f"MLP(input_dim={self.input_dim}, output_dim={self.output_dim}, "
            f"hidden_dim={self.hidden_dim})"
        )


class PrefixTuning(_nn.Module):
    """Class that generates prefix tensors for key and value."""

    def __init__(
        self,
        embedding_dim: int,
        prefix_length: int,
        prefix_dim: int = None,
        prefix_hidden_dim: int = 512,
        dropout_p: float = 0,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.prefix_length = prefix_length
        self.prefix_dim = prefix_dim if prefix_dim else embedding_dim
        self.prefix_hidden_dim = prefix_hidden_dim
        self.dropout_p = dropout_p
        self.prefix_embedding = _nn.Parameter(
            _torch.empty(
                self.prefix_length,
                self.prefix_dim,
                device=device,
                dtype=dtype,
            )
        )

        self.prefix_k_proj = MLP(
            input_dim=self.prefix_dim,
            output_dim=self.embedding_dim,
            hidden_dim=self.prefix_hidden_dim,
            device=device,
            dtype=dtype,
        )
        self.prefix_v_proj = MLP(
            input_dim=self.prefix_dim,
            output_dim=self.embedding_dim,
            hidden_dim=self.prefix_hidden_dim,
            device=device,
            dtype=dtype,
        )

    def forward(self, x: _torch.Tensor) -> Tuple[_torch.Tensor, ...]:
        predix_k = self.prefix_k_proj(self.prefix_embedding)
        predix_v = self.prefix_v_proj(self.prefix_embedding)
        return _torch.broadcast_to(predix_k, (x.shape[0], -1, -1)), _torch.broadcast_to(
            predix_v, (x.shape[0], -1, -1)
        )

    def extra_repr(self) -> str:
        return (
            f"PrefixTuning(prefix_dim={self.prefix_dim}, "
            f"embedding_dim={self.embedding_dim}, "
            f"prefix_length={self.prefix_length}, "
            f"prefix_hidden_dim={self.prefix_hidden_dim}, "
            f"dropout_p={self.dropout_p})"
        )


@_attach_config_class
class PrefixTuningAttention(_LayerAdapter):
    """PrefixTuning Adapter class which prepend prefix to key and value."""

    @property
    def prefix_length(self) -> int:
        return self._prefix_length

    @property
    def prefix_hidden_dim(self) -> int:
        return self._prefix_hidden_dim

    def __init__(
        self,
        prefix_length: int,
        embedding_dim: int,
        prefix_dim: int = None,
        prefix_hidden_dim: int = 512,
        dropout_p: float = 0,
        device: _Optional[_Union[_torch.device, str]] = None,
        dtype: _Optional[_Union[_torch.dtype, str]] = _torch.float32,
    ) -> None:
        super().__init__()
        self._prefix_length = prefix_length
        self._prefix_hidden_dim = prefix_hidden_dim

        self.prefix_tuning = PrefixTuning(
            embedding_dim=embedding_dim,
            prefix_length=prefix_length,
            prefix_dim=prefix_dim,
            prefix_hidden_dim=prefix_hidden_dim,
            dropout_p=dropout_p,
            device=device,
            dtype=dtype,
        )

    def _transform_inputs(self, *, args, kwargs):
        return args, kwargs

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _torch.Tensor,
    ):
        if not len(args) == 1:
            raise ValueError(
                "Input to PrefixTuning adapter should be single torch tensor object."
            )

        prefix_k, prefix_v = self.prefix_tuning(args[0])
        k_adapted = _torch.concat([prefix_k, outputs[1]], dim=1)
        v_adapted = _torch.concat([prefix_v, outputs[2]], dim=1)
        return outputs[0], k_adapted, v_adapted


@_attach_config_class
class PrefixTuningALiBiMask(_LayerAdapter):
    """Class that generates modified ALiBi mask based on prefix_lengh."""

    def __init__(
        self,
        prefix_length: int,
    ) -> None:
        super().__init__()
        self.prefix_length = prefix_length

    def extra_repr(self) -> str:
        return f"PrefixTuningALiBiMask(prefix_length={self.prefix_length})"

    def _transform_inputs(self, *, args, kwargs):
        """
        In tamm's ALiBi implementation, segment_ids are used to compute causal mask.
        We add the segment_ids based on prefix_length to generate the correct mask.
        For detail ALiBi implementation, please refer to tamm.layers.attention_mask.py.
        """
        if len(args) != 1:
            raise ValueError(
                "Input to PrefixTuning adapter should be single torch tensor"
            )
        inputs = args[0]
        prefix_segment_ids = (
            _torch.ones(inputs.shape[0], self.prefix_length).to(device=inputs.device)
            * inputs[0, 0]
        )
        full_segment_ids = _torch.concat([prefix_segment_ids, inputs], dim=1)
        return full_segment_ids

    def _transform_outputs(
        self,
        *,
        args: _Any,
        kwargs: _Any,
        transformed_args: _Any,
        transformed_kwargs: _Any,
        outputs: _torch.Tensor,
    ):
        return outputs[:, :, self.prefix_length :].contiguous()



"""
Command for showing the model config when given its model id as input.
"""
import logging
import pprint
import warnings

import tamm
from tamm import _helpers
from tamm._adapters_v1.utils import get_num_adapter_params
from tamm._cli.common import argument, command
from tamm.utils import torch_utils as _torch_utils

logger = logging.getLogger(__name__)
# pylint: disable=redefined-builtin


@command("show", help_text="Print a tamm config")
@argument(
    "name",
    metavar="<OBJECT_ID>",
    type=str,
    help=(
        "Any identifiers recognized by tamm--use ``tamm ls models -l`` "
        "or ``tamm ls preprocessors``, etc., for available options"
    ),
)
def show_object(name):
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            config = tamm.create_model_config(name)
        print(config)
        with _helpers.catch_and_log_exceptions(context_name="print model size"):
            _print_model_size(config)
        return
    except Exception:
        logger.debug("%s does not exist in model registry", name)

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            config = tamm.create_tokenizer_config(name)
        print(config)
        return
    except Exception:
        logger.debug("%s does not exist in tokenizer registry", name)

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            config = tamm.preprocessors.describe(name)
        pprint.pprint(config, compact=True, width=100)
        return
    except Exception:
        logger.debug("%s does not exist in preprocessors registry", name)

    print(f"{name} is neither a recognized model id, tokenizer id, nor preprocessor id")


def _print_model_size(config):
    model = config.create_model(device="meta")
    num_params = _torch_utils.get_num_params(model)
    num_adapter_params = get_num_adapter_params(model)
    print(f"Parameter count: {num_params}")
    print(f"Adapter parameter count: {num_adapter_params}")

    num_bytes = num_params * 2
    num_bytes = _helpers.size_in_bytes_to_string(num_bytes)
    print(f"Model size (fp16): {num_bytes}")



"""
utils.user_dir_utils
====================

This module implements helper functions for working with |tamm|'s user dir.

.. autofunction:: tamm.utils.user_dir_utils.get_tamm_user_dir

.. autofunction:: tamm.utils.user_dir_utils.get_tamm_cache_dir

.. autofunction:: tamm.utils.user_dir_utils.get_tamm_lock_dir

.. autofunction:: tamm.utils.user_dir_utils.get_tamm_tmp_dir
"""

import os as _os

from tamm.runtime_configuration import rc as _rc


def get_tamm_user_dir() -> str:
    """
    Returns |tamm|'s user dir, creating the directory if it does not
    yet exist.  This path may be configured using :attr:`tamm.rc.user_dir`
    or the ``TAMM_USER_DIR`` env var.
    """
    return _maybe_make_dir(_rc.user_dir)


def get_tamm_cache_dir() -> str:
    """
    Returns |tamm|'s cache dir, creating the directory if it does not
    yet exist.  The cache dir is a local directory for caching downloads,
    and it is a subdirectory of the user dir.
    """

    result = _os.path.join(get_tamm_user_dir(), "cache")
    return _maybe_make_dir(result)


def get_tamm_lock_dir() -> str:
    """
    Returns |tamm|'s lock dir, creating the directory if it does not
    yet exist.  The lock dir is a local directory for |tamm| file locks,
    and it is a subdirectory of the user dir.
    """

    result = _os.path.join(get_tamm_user_dir(), "lock")
    return _maybe_make_dir(result)


def get_tamm_tmp_dir() -> str:
    """
    Returns |tamm|'s tmp dir, creating the directory if it does not
    yet exist.  The tmp dir is a local directory for |tamm| tmp files,
    and it is a subdirectory of the user dir.
    """

    result = _os.path.join(get_tamm_user_dir(), "tmp")
    return _maybe_make_dir(result)


def _maybe_make_dir(dirpath: str) -> str:
    if not _os.path.exists(dirpath):
        _os.makedirs(dirpath, exist_ok=True)
    return dirpath



import torch as _torch
import torch.nn as _nn

from tamm.layers.common import LayerMixin as _LayerMixin


class FlattenedCrossEntropyLoss(_nn.CrossEntropyLoss, _LayerMixin):
    # pylint: disable=redefined-builtin
    def forward(self, input: _torch.Tensor, target: _torch.Tensor) -> _torch.Tensor:
        input = input.view(-1, input.shape[-1])
        target = target.flatten()
        return super().forward(input, target)



from typing import Tuple as _Tuple
from typing import Union as _Union

import torch as _torch

from tamm.layers.transformer.kv_cache import BaseKVCacheView as _BaseKVCacheView


class HFStyleKVCacheView(_BaseKVCacheView):
    def __init__(
        self,
        past_key_values: _Tuple[_Tuple[_torch.Tensor]],
        segment_ids: _torch.Tensor,
        positions: _torch.Tensor,
    ):
        self._past_key_values = list(past_key_values)
        self._segment_ids = segment_ids
        self._positions = positions

    @classmethod
    def create_empty_cache(cls, *, num_layers, segment_ids, positions):
        past_key_values = tuple(tuple() for _ in range(num_layers))
        return cls(past_key_values, segment_ids=segment_ids, positions=positions)

    @property
    def past_key_values(self):
        return self._past_key_values

    @property
    def _is_empty(self):
        if len(self._past_key_values) == 0:
            return True
        return len(self._past_key_values[0]) == 0

    @property
    def _first_tensor(self):
        """
        Returns the first tensor in the cache.  Raises an :obj:`IndexError`
        if the cache is empty.
        """
        return self._past_key_values[0][0]

    @property
    def num_layers(self) -> int:
        return len(self._past_key_values)

    @property
    def length(self) -> int:
        if self._is_empty:
            return 0
        return self._first_tensor.size(2)

    @property
    def dtype(self) -> _Union[_torch.dtype, None]:
        if self._is_empty:
            return None
        return self._first_tensor.dtype

    @property
    def device(self) -> _torch.device:
        return self._segment_ids.device

    def write_to_tail(
        self, key: _torch.Tensor, value: _torch.Tensor, *, layer_idx: int
    ) -> None:
        past_kv = self._past_key_values[layer_idx]
        if len(past_kv) == 0:
            new_key = key
            new_value = value
        else:
            past_key, past_value = past_kv
            new_key = _torch.cat([past_key, key], dim=1)
            new_value = _torch.cat([past_value, value], dim=1)
        self._past_key_values[layer_idx] = (new_key, new_value)

    def write_segment_ids_to_tail(self, segment_ids: _torch.Tensor) -> None:
        """
        This class assumes all segment ids are defined during __init__(),
        so we ignore the segment IDs the that the tamm model passes during
        forward() (they should be redundant).
        """

    def write_positions_to_tail(self, positions: _torch.Tensor) -> None:
        """
        This class assumes all positions are defined during __init__(),
        so we ignore the positions that the tamm model passes during
        forward() (they should be redundant).
        """

    def read(self, layer_idx: int) -> _Tuple[_torch.Tensor, _torch.Tensor]:
        return self._past_key_values[layer_idx]

    def read_segment_ids(self) -> _torch.Tensor:
        return self._segment_ids

    def read_positions(self) -> _torch.Tensor:
        return self._positions
